import os
from datetime import datetime

# Custom log function
def log_command(message, log_file="log.log"):
    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    log_entry = f"{timestamp} - {message}\n"

    # Append log entry to the log file
    with open(log_file, "a") as f:
        f.write(log_entry)

    print(log_entry.strip())  # Optional: print to the console



from prolysis.util.functions import n_edges
from prolysis.util.functions import add_SE
from prolysis.discovery.candidate_search.is_allowed_2 import is_allowed
import networkx as nx


def adj(node_set, net,n_thr):
    adj_set = set()
    for node in node_set:
        if len(set(net.neighbors(node)))>0:
            max_out = max([n_edges(net, {node}, {x}) for x in net.neighbors(node)])
            filt_num = max_out * n_thr
            adj_set = adj_set.union(set([x for x in net.neighbors(node) if n_edges(net, {node}, {x})>=filt_num]))
        else:
            adj_set = adj_set.union(set(net.neighbors(node)))
    return adj_set


def find_possible_partitions(net, rules, st_net, en_net,fp):
    activity_list = set(net.nodes) - {'start', 'end'}
    queue = [(set(), {'start'})]
    visited = set()
    valid = []
    reserve = []
    low_pen = 10000
    if n_edges(net, {'start'}, {'end'}) > 0:
        # na, block, exclude_dic = is_allowed(activity_list, set(), rules, st_net, en_net)
        if len(rules[1]) > 0:
            na, block, penalty = is_allowed(activity_list, set(), rules)
        else:
            na = set()
            penalty = {'seq': 0, 'exc': 0, 'par': 0, 'loop': 0, 'loop_tau': 0, 'exc_tau':0}
            block = False
        possible_cuts = {'exc_tau'} - na
        reserve += [(activity_list, set(), ct, penalty[ct]) for ct in possible_cuts]

        # if 'exc_tau' not in na:
        #     valid.append((activity_list, set(), {'exc_tau'}))
        # else:
        #     reserve.append(((activity_list, set(), {'exc_tau'}), 'exc_tau', sum([y[2] for y in exclude_dic['exc_tau']])))

    if (n_edges(net, set(en_net.keys()), set(st_net.keys())) > 0) and set(st_net.keys())==set([x for x in st_net.keys() if (x,x) in fp.edges() and fp[x][x]['weight'] > 0]):
        # na, block, exclude_dic = is_allowed(activity_list, set(), rules, st_net, en_net)
        if len(rules[1]) > 0:
            na, block, penalty = is_allowed(activity_list, set(), rules)
        else:
            na = set()
            penalty = {'seq': 0, 'exc': 0, 'par': 0, 'loop': 0, 'loop_tau': 0, 'exc_tau':0}
            block = False
        possible_cuts = {'loop_tau'} - na
        reserve += [(activity_list, set(), ct, penalty[ct]) for ct in possible_cuts]

        # if 'loop_tau' not in na:
        #     valid.append((activity_list, set(), {'loop_tau'}))
        # else:
        #     reserve.append(((activity_list, set(), {'loop_tau'}), 'loop_tau', sum([y[2] for y in exclude_dic['loop_tau']])))

    while len(queue) != 0:
        current = queue.pop()
        current_set, current_adj = current

        for x in current_adj:
            new_state = current_set | {x}
            new_state = add_SE(new_state, st_net, en_net)

            if frozenset(new_state) not in visited:
                new_adj = (current_adj | adj({x}, net,0)) - new_state
                visited.add(frozenset(new_state))
                # possible_cuts = set()

                B = activity_list - new_state
                # na, block, exclude_dic = is_allowed(new_state, B, rules, st_net, en_net)

                if (len(B) == 0) or (len(B) == len(activity_list)):
                    queue.append((new_state, new_adj))
                    continue

                B = add_SE(B,st_net,en_net)
                netA = net.subgraph(new_state)
                netB = net.subgraph(B)

                # 'start' ~> netB
                if 'start' in netB:
                    not2startB = set(netB.nodes) - set(nx.descendants(netB, 'start')) - {'start', 'end'}
                else:
                    not2startB = set(netB.nodes) - {'start', 'end'}
                # 'end' ~> netA
                if 'end' in netA:
                    not2endA = set(netA.nodes) - set(nx.ancestors(netA, 'end')) - {'start', 'end'}
                else:
                    not2endA = set(netA.nodes) - {'start', 'end'}
                # 'end' ~> netB
                if 'end' in netB:
                    not2endB = set(netB.nodes) - set(nx.ancestors(netB, 'end')) - {'start', 'end'}
                else:
                    not2endB = set(netB.nodes) - {'start', 'end'}


                possible_cuts = set()
                if not not2endB:
                    possible_cuts.add("seq")
                    if len(not2endA) == 0:
                        if n_edges(net, B-{'start','end'}, new_state & set(st_net.keys())) != 0 and n_edges(net, new_state & set(en_net.keys()), B-{'start','end'}) != 0:
                                possible_cuts.add("loop")
                        if len(not2startB) == 0 and (B not in visited):
                            possible_cuts.update(["exc", "par"])
                elif not not2endA:
                    if n_edges(net, B, new_state & set(st_net.keys())) != 0 and n_edges(net,new_state & set(en_net.keys()),B) != 0:
                            possible_cuts.add("loop")
                # for x in possible_cuts:
                #     if x in na:
                #         reserve.append(((new_state, B, possible_cuts),x, sum([y[2] for y in exclude_dic[x]])))
                # possible_cuts = possible_cuts - na

                if len(rules[1]) > 0:
                    na, block, penalty = is_allowed(new_state, B, rules)
                else:
                    na = set()
                    penalty = {'seq': 0, 'exc': 0, 'par': 0, 'loop': 0, 'loop_tau': 0, 'exc_tau': 0}
                    block = False
                reserve += [(new_state, B, ct, penalty[ct]) for ct in possible_cuts]
                possible_cuts = possible_cuts - na
                if len(possible_cuts) > 0:
                    valid.append((new_state, B, possible_cuts))

                if not block:
                    queue.append((new_state, new_adj))
    # print(penalty)
    min_value = min(reserve, key=lambda x: x[3])[3]
    if min_value>0:
        print(f"warning! the minimum cost is {min_value} and not 0")
    result = [(tup[0],tup[1],tup[2]) for tup in reserve if tup[3] == min_value]
    # return valid,reserve
    return result,reserve


import time


def snake_to_camel(input: str) -> str:
    # Can swap out with more sophisticated implementation if needed
    camel_cased = "".join(x.capitalize() for x in input.lower().split("_"))
    if camel_cased:
        return camel_cased[0].lower() + camel_cased[1:]
    else:
        return camel_cased


def time_it(func): 
    '''Decorator that reports the execution time.'''
  
    def wrap(*args, **kwargs): 
        start_time = time.perf_counter()
        result = func(*args, **kwargs) 
        end_time = time.perf_counter()
          
        print(func.__name__, end_time - start_time) 
        return result 
    return wrap 


from prolysis.util.functions import n_edges
from prolysis.discovery.candidate_search.is_allowed_2 import is_allowed_single
def check_base_case(rules ,sup_thr, ratio,self):
    netP = self.netP
    activitiesP = set(netP.nodes) - {'start', 'end'}

    if not self.ignore_M:
        netM = self.netM
        activitiesM = set(netM.nodes) - {'start', 'end'}
        if len(activitiesP)>1:
            base_check = False
            cut = "not_base"
        elif len(activitiesP)==0:
            base_check = True
            self.detected_cut = 'empty_log'
            cut = ('none', 'empty_log', 'none', 'none')
        else:
            base_check = True
            act = activitiesP.pop()
            na, penalty = is_allowed_single(act, rules)
            #if absent we do not want to hide a transition!
            if na == {'single_single', 'xor_single', 'loop_single'}:
                # self.detected_cut = 'empty_log'
                # cut = ('none', 'empty_log', 'none', 'none')
                print('na includes all in base case check')

            cP = max(0, sup_thr * n_edges(netP,{'start'},{act, 'end'})-n_edges(netP,{'start'}, {'end'}))
            cM = max(0,sup_thr * n_edges(netM,{'start'},{act, 'end'})-n_edges(netM,{'start'}, {'end'}))
            cost_single_exc =  cP - ratio * self.size_adj * cM
            if (cost_single_exc <= 0) and n_edges(netP,{'start'}, {'end'}) > 0 and 'xor_single' not in na:
                cut = ((({act}, set()), 'exc2', cP, cM,cost_single_exc,{str(act): {'missing': cP, 'deviating': 0}},{str(act): {'missing': cM, 'deviating': 0}}))
            else:
                clP = max(0,sup_thr* ((n_edges(netP,{'start',act},{act}))/2)-n_edges(netP,{act},{act}))

                if len(activitiesM) > 0:
                    clM = max(0, sup_thr * ((n_edges(netM, {'start', act}, {act})) / 2) - n_edges(netM, {act}, {act}))
                else:
                    clM = max(0, sup_thr * ((n_edges(netM, {'start', act}, {act})) / 2))
                cost_single_loop = clP - ratio * self.size_adj * clM
                if (cost_single_loop <= 0) and n_edges(netP,{act},{act})>0 and 'loop_single' not in na:
                    cut = (({act}, set()), 'loop1', 'none', 'none')
                else:
                    # single activity
                    self.detected_cut = 'single_activity'
                    cut =((
                    ({act}, set()), 'single_activity', n_edges(netP,{'start'}, {'end'}), n_edges(netM,{'start'}, {'end'}),
                    n_edges(netP,{'start'}, {'end'}) - ratio * self.size_adj * n_edges(netM,{'start'}, {'end'}),
                    {str(act): {'deviating': n_edges(netP,{'start'}, {'end'}), 'missing': 0}},
                    {str(act): {'deviating': n_edges(netM,{'start'}, {'end'}), 'missing': 0}}))
        return base_check, cut
    else:
        if len(activitiesP) > 1:
            base_check = False
            cut = "not_base"
        elif len(activitiesP) == 0:
            base_check = True
            self.detected_cut = 'empty_log'
            cut = ('none', 'empty_log', 'none', 'none')
        else:
            base_check = True
            act = activitiesP.pop()
            na, penalty = is_allowed_single(act, rules)
            # if absent we do not want to hide a transition!
            if na == {'single_single', 'xor_single', 'loop_single'}:
                # self.detected_cut = 'empty_log'
                # cut = ('none', 'empty_log', 'none', 'none')
                print('na includes all in base case check')

            cP = max(0, sup_thr * n_edges(netP, {'start'}, {act, 'end'}) - n_edges(netP, {'start'}, {'end'}))
            cost_single_exc = cP
            if (cost_single_exc <= 0) and n_edges(netP, {'start'}, {'end'}) > 0 and 'xor_single' not in na:
                cut = ((({act}, set()), 'exc2', cP, 0, cost_single_exc, {str(act): {'missing': cP, 'deviating': 0}},
                        {str(act): {'missing': 0, 'deviating': 0}}))
            else:
                clP = max(0, sup_thr * ((n_edges(netP, {'start', act}, {act})) / 2) - n_edges(netP, {act}, {act}))
                cost_single_loop = clP
                if (cost_single_loop <= 0) and n_edges(netP, {act}, {act}) > 0 and 'loop_single' not in na:
                    cut = (({act}, set()), 'loop1', 'none', 'none')
                else:
                    # single activity
                    self.detected_cut = 'single_activity'
                    cut = ((
                        ({act}, set()), 'single_activity', n_edges(netP, {'start'}, {'end'}),
                        0,
                        n_edges(netP, {'start'}, {'end'}),
                        {str(act): {'deviating': n_edges(netP, {'start'}, {'end'}), 'missing': 0}},
                        {str(act): {'deviating': 0, 'missing': 0}}))
        return base_check, cut





import prolysis.rules_handling.declare_processing as declare_processing


def is_allowed(activity,rules,low_pen):
    penalty = {'single':0, 'xor':0, 'loop':0}

    da_list = rules[1]
    S_mapped = rules[2]
    exclude = []


    for cut_type in {'single', 'xor', 'loop'}:
        for dfa in da_list:
            if (declare_processing.check_all(S1,S2,S1&{x for x in dfa[0][1]}, S2&{x for x in dfa[0][1]}, dfa[1], cut_type, S_mapped)):
                print(dfa[0])
                exclude.append(cut_type)
                if penalty[cut_type]+dfa[2]>low_pen:
                    penalty[cut_type] += dfa[2]
                    break
                else:
                    penalty[cut_type] += dfa[2]
                    low_pen = min(low_pen,penalty[cut_type])
    # if sum(penalty.values())>0:
    #     print(penalty)
    print(low_pen)
    return set(exclude), set(), penalty,low_pen



import prolysis.discovery.split_functions.split as split
from prolysis.discovery.candidate_search.search import find_possible_partitions
from prolysis.discovery.base_case.check import check_base_case
from prolysis.discovery.cut_quality.cost_functions import cost_functions
from prolysis.util.functions import generate_nx_graph_from_log, generate_nx_indirect_graph_from_log, get_input_output_B_indices
from prolysis.util.functions import convert_activities_to_array as act2ary
from prolysis.discovery.cut_quality.cost_functions.cost_functions import overal_cost
import networkx as nx
import numpy as np
from collections import Counter


class SubtreePlain(object):
    def __init__(self, logp,logm, rec_depth, noise_threshold= None,
                   parameters=None, sup= None, ratio = None, size_par = None, rules = None):

        nt = 0.000 * sum(logp.values())
        self.rec_depth = rec_depth
        self.noise_threshold = noise_threshold
        self.log = logp

        if sum(self.log.values())==0:
            self.activitiesP = set()
            self.start_activitiesP = {}
            self.end_activitiesP = {}
        else:
            self.activitiesP = set(y for x in logp for y in x if x!=())
            self.start_activities = set(x[0] for x in logp if x!=())
            self.end_activities = set(x[-1] for x in logp if x!=())


        if not logm:
            self.ignore_M = True
            self.activitiesM = set()
            self.start_activitiesM = {}
            self.end_activitiesM = {}
            self.size_adj = 1
        else:
            self.ignore_M = False
            self.logM = logm
            self.activitiesM = set(y for x in logm for y in x if x!=())
            self.start_activitiesM = set(x[0] for x in logm if x!=())
            self.end_activitiesM = set(x[-1] for x in logm if x!=())


        self.netP = generate_nx_graph_from_log(self.log, nt, self.activitiesM)
        self.all_acts = self.activitiesP.union(self.activitiesM)
        self.nodes_order = list(self.all_acts - {'start', 'end'})
        self.nodes_order.append('start')
        self.nodes_order.append('end')
        self.mapping = {i: activity for i, activity in enumerate(self.nodes_order)}
        self.mapping_rev = {v: k for k, v in self.mapping.items()}
        self.adj_matrixP = nx.to_numpy_array(self.netP, nodelist=self.nodes_order)  # Ensure correct ordering
        self.adj_dict = {n: set(self.netP.neighbors(n)) for n in self.netP.nodes}

        if logm:
            self.netM = generate_nx_graph_from_log(self.logM,nt,self.activitiesP)
            self.adj_matrixM = nx.to_numpy_array(self.netM, nodelist=self.nodes_order)
            self.size_adj = sum(self.log.values())/sum(self.logM.values())


        self.original_log = logp
        self.parameters = parameters
        self.detected_cut = None
        self.children = []
        self.detect_cut(parameters=parameters, sup= sup, ratio = ratio, size_par = size_par, rules = rules)


    def detect_cut(self, parameters=None, sup= None, ratio = None, size_par = None, rules = None):

        if parameters is None:
            parameters = {}

        # check base cases:
        isbase, cut = check_base_case(rules , sup, ratio, self)

        if isbase==False:
            fP = generate_nx_indirect_graph_from_log(self.log,self.activitiesM)
            fp_adj = nx.to_numpy_array(fP, nodelist=self.nodes_order[:-2])
            if not self.ignore_M:
                fM = generate_nx_indirect_graph_from_log(self.logM,self.activitiesP)
                fm_adj = nx.to_numpy_array(fM, nodelist=self.nodes_order[:-2])

            possible_partitions, reserve2 = find_possible_partitions(rules, self.start_activities,
                                                                    self.end_activities, fP, self.adj_matrixP, self.nodes_order,self.mapping_rev,self.adj_dict)

            cut = []

            ratio_backup = ratio

            min_cost = 65534
            for pp in possible_partitions:
                A = pp[0] - {'start', 'end'}
                A_np = np.array([self.mapping_rev[s] for s in A], dtype=np.int16)
                B = pp[1] - {'start', 'end'}
                B_np = np.array([self.mapping_rev[s] for s in B], dtype=np.int16)

                start_A_P = self.start_activities & A
                end_A_P = self.end_activities & A
                input_B_P, output_B_P = get_input_output_B_indices(self.adj_matrixP,B_np)

                if not self.ignore_M:
                    start_A_M = self.start_activitiesM & A
                    end_A_M = self.end_activitiesM & A
                    input_B_M, output_B_M = get_input_output_B_indices(self.adj_matrixM, B_np)
                    if sum(self.logM.values()) == 0:
                        ratio = 0
                    else:
                        ratio = ratio_backup

                type = pp[2]


                #####################################################################
                # seq check
                if type=="seq":
                    min_cost, dev, mis, valid = cost_functions.cost_seq(self.adj_matrixP, A_np,
                    B_np, sup, fp_adj, min_cost, self.ignore_M)
                    cost_seq_P_dict = {'missing':mis, 'deviating':dev}
                    cost_seq_P = mis+dev
                    if not self.ignore_M:
                        min_cost, dev, mis, valid = cost_functions.cost_seq(self.adj_matrixM, A_np,
                                                                  B_np, sup, fm_adj, min_cost, self.ignore_M)
                        cost_seq_M_dict = {'missing':mis, 'deviating':dev}
                        cost_seq_M = mis+dev
                    else:
                        cost_seq_M = 0
                        cost_seq_M_dict = {}
                    if valid:
                        cut.append(((A, B), 'seq', cost_seq_P, cost_seq_M, overal_cost(cost_seq_P,cost_seq_M,ratio,self.size_adj),cost_seq_P_dict,cost_seq_M_dict))
                #####################################################################
                # xor check
                if type=="exc":
                    min_cost, dev, valid = cost_functions.cost_exc(self.adj_matrixP, A_np,
                     B_np,min_cost, self.ignore_M)
                    cost_exc_P_dict = {'missing':0, 'deviating':dev}
                    cost_exc_P = dev

                    if not self.ignore_M:
                        min_cost, dev, valid = cost_functions.cost_exc(self.adj_matrixM,A_np,
                                                                  B_np, min_cost, self.ignore_M)
                        cost_exc_M_dict = {'missing':0, 'deviating':dev}
                        cost_exc_M = dev
                    else:
                        cost_exc_M_dict = {}
                        cost_exc_M = 0
                    if valid:
                        cut.append(((A, B), 'exc', cost_exc_P, cost_exc_M, overal_cost(cost_exc_P,cost_exc_M,ratio,self.size_adj),cost_exc_P_dict,cost_exc_M_dict))
                #####################################################################
                # xor-tau check
                if type=="exc_tau":
                    mis = cost_functions.cost_exc_tau(self.adj_matrixP,sup,self.mapping_rev['start'],self.mapping_rev['end'])
                    cost_exc_tau_P_dict = {'missing':mis, 'deviating':0}
                    cost_exc_tau_P = mis
                    if not self.ignore_M:
                        mis = cost_functions.cost_exc_tau(self.adj_matrixM, sup,self.mapping_rev['start'],self.mapping_rev['end'])
                        cost_exc_tau_M_dict = {'missing':mis, 'deviating':0}
                        cost_exc_tau_M = mis
                    else:
                        cost_exc_tau_M_dict = {}
                        cost_exc_tau_M = 0
                    cut.append(((A.union(B), set()), 'exc_tau',cost_exc_tau_P , cost_exc_tau_M, overal_cost(cost_exc_tau_P,cost_exc_tau_M,ratio,self.size_adj),cost_exc_tau_P_dict,cost_exc_tau_M_dict))
                #####################################################################
                # parallel check
                if type=="par":
                    min_cost, mis, valid = cost_functions.cost_par(self.adj_matrixP, A_np,
                     B_np, sup, min_cost,self.ignore_M)
                    cost_par_P_dict = {'missing':mis, 'deviating':0}
                    cost_par_P = mis
                    if not self.ignore_M:
                        min_cost, mis, valid = cost_functions.cost_par(self.adj_matrixM, A_np,
                                                                  B_np, sup, min_cost,self.ignore_M)
                        cost_par_M_dict = {'missing':mis, 'deviating':0}
                        cost_par_M = mis
                    else:
                        cost_par_M_dict = {}
                        cost_par_M = 0
                    if valid:
                        cut.append(((A, B), 'par', cost_par_P, cost_par_M, overal_cost(cost_par_P,cost_par_M,ratio,self.size_adj),cost_par_P_dict,cost_par_M_dict))
                #####################################################################
                # loop check
                if type=="loop":
                    min_cost,dev, mis, valid = cost_functions.cost_loop(self.adj_matrixP, A_np, B_np, sup,
                                                                   act2ary(start_A_P, self.mapping_rev), act2ary(end_A_P, self.mapping_rev), input_B_P,
                                                                   output_B_P,self.mapping_rev['start'],self.mapping_rev['end'],min_cost,self.ignore_M)

                    cost_loop_P_dict = {'missing':mis, 'deviating':dev}
                    cost_loop_P = mis+dev
                    if not self.ignore_M:
                        min_cost,dev, mis, valid = cost_functions.cost_loop(self.adj_matrixM,
                                                                       A_np,
                                                                       B_np,
                                                                       sup,
                                                                       act2ary(start_A_M,
                                                                                                   self.mapping_rev),
                                                                       act2ary(end_A_M,
                                                                                                   self.mapping_rev),
                                                                       input_B_M,
                                                                       output_B_M,
                                                                       self.mapping_rev['start'], self.mapping_rev['end'],min_cost,self.ignore_M)
                        cost_loop_M_dict = {'missing':mis, 'deviating':dev}
                        cost_loop_M = mis+dev
                    else:
                        cost_loop_M_dict = {}
                        cost_loop_M = 0

                    if valid:
                        cut.append(((A, B), 'loop', cost_loop_P, cost_loop_M, overal_cost(cost_loop_P,cost_loop_M,ratio,self.size_adj),cost_loop_P_dict,cost_loop_M_dict))

                if type=="loop_tau":
                    mis = cost_functions.cost_loop_tau(self.adj_matrixP, sup, act2ary(start_A_P, self.mapping_rev),act2ary(end_A_P, self.mapping_rev),self.mapping_rev['start'],self.mapping_rev['end'])
                    cost_loop_P_dict = {'missing':mis, 'deviating': 0 }
                    cost_loop_P = mis
                    if not self.ignore_M:
                        mis = cost_functions.cost_loop_tau(self.adj_matrixM, sup, act2ary(start_A_M, self.mapping_rev),act2ary(end_A_M, self.mapping_rev),self.mapping_rev['start'],self.mapping_rev['end'])
                        cost_loop_M_dict = {'missing':mis, 'deviating': 0 }
                        cost_loop_M = mis
                    else:
                        cost_loop_M_dict = {}
                        cost_loop_M = 0
                    cut.append(((self.start_activities, self.end_activities), 'loop_tau', cost_loop_P,cost_loop_M, overal_cost(cost_loop_P,cost_loop_M,ratio,self.size_adj), cost_loop_P_dict,cost_loop_M_dict))


            if not cut:
                print("no good cut exists")
            sorted_cuts = sorted(cut, key=lambda x: (x[4], x[2],['exc_tau','exc','seq','par','loop','loop_tau'].index(x[1]), abs(len(x[0][0])-len(x[0][1]))))
            cut = sorted_cuts[0]

        # print(cut[:-2])

        map_cut_op = {'par': 'parallel', 'seq': 'sequential', 'exc': 'xor', 'exc_tau':'xor', 'exc2': 'xor',
                      'loop': 'loopCut', 'loop1': 'loopCut', 'loop_tau': 'loopCut'}



        if cut[1] in map_cut_op.keys():
            self.detected_cut = map_cut_op[cut[1]]
            LAP, LBP = split.split(cut[1], [cut[0][0], cut[0][1]], self.log)
            if not self.ignore_M:
                LAM, LBM = split.split(cut[1], [cut[0][0], cut[0][1]], self.logM)
            else:
                LAM = Counter()
                LBM = Counter()
            new_logs = [[LAP, LAM], [LBP, LBM]]
            for l in new_logs:
                self.children.append(
                    SubtreePlain(l[0], l[1],
                                 self.rec_depth + 1,
                                 noise_threshold=self.noise_threshold,
                                 parameters=parameters, sup=sup, ratio=ratio, size_par=size_par, rules=rules))
        elif cut[1]!='single_activity' and cut[1]!='empty_log':
            print('It should not happen, if you see this error there could be a bug in the code!')




import networkx as nx
import json
import numpy as np
from numba import njit
from numba.typed import List


# @njit
def select_submatrix(adj_matrix, nodes_order, included_nodes):
    # Select submatrix
    new_adj_matrix = adj_matrix[included_nodes[:, None], included_nodes]

    return new_adj_matrix



@njit
def bfs_descendants(adj_matrix, start_index):
    n = adj_matrix.shape[0]

    # Using Numba-typed list for performance
    queue = List()
    queue.append(start_index)

    visited = np.zeros(n, dtype=np.uint8)
    visited[start_index] = 1

    # Use a fixed-size NumPy array (worst-case: all nodes reachable)
    descendants = np.full(n, -1, dtype=np.int32)  # Initialize with -1 (invalid index)
    count = 0  # Track number of valid descendants

    while len(queue) > 0:
        node = queue.pop(0)
        for neighbor in range(n):
            if adj_matrix[node, neighbor] > 0 and not visited[neighbor]:
                queue.append(neighbor)
                visited[neighbor] = 1
                descendants[count] = neighbor
                count += 1  # Move to next slot

    return descendants[:count]  # Return only valid indices


@njit
def bfs_ancestors(adj_matrix, start_index):
    return bfs_descendants(adj_matrix.T, start_index)  # BFS on reversed edges




def n_edges(net, S, T):
    edges_reweight = list(nx.edge_boundary(net, S, T, data='weight', default=1))
    return sum(weight for u, v, weight in edges_reweight if (u in S and v in T))


# def n_edges2(adj_matrix, index_to_activity, S, T):
#     # Convert activity names to row indices
#     S_indices = {i for i, activity in index_to_activity.items() if activity in S}
#     T_indices = {i for i, activity in index_to_activity.items() if activity in T}
#
#     # Sum up the edges from S to T in the adjacency matrix
#     edge_count = sum(adj_matrix[i, j] for i in S_indices for j in T_indices)
#
#     return edge_count

# @njit
def sum_out_degree_numba(adj_matrix, A_indices):
    total_degree = 0.0
    for i in A_indices:
        total_degree += adj_matrix[i, :].sum()
    return total_degree


# @njit
def sum_out_degree_single(adj_matrix, A_index):
    return adj_matrix[A_index, :].sum()

# @njit
def get_edge_weight_numba(adj, node1_idx, node2_idx):
    return adj[node1_idx, node2_idx]

# @njit
def n_edges_numba(adj, S_indices, T_indices):
    count = 0
    for s in S_indices:
        for t in T_indices:
            count += adj[s, t]
    return count


def get_edge_weight(graph, node1, node2):
    edge_data = graph.get_edge_data(node1, node2)
    if edge_data is None:
        return 0
    else:
        return edge_data['weight']

def get_edge_weight2(adj_matrix, index_to_activity, node1, node2):
    # Convert node names to indices
    node1_index = next((i for i, act in index_to_activity.items() if act == node1), None)
    node2_index = next((i for i, act in index_to_activity.items() if act == node2), None)

    # If nodes are not found, return 0
    if node1_index is None or node2_index is None:
        return 0

    # Retrieve the edge weight from the adjacency matrix
    return adj_matrix[node1_index, node2_index]

def sum_out_degree(adj_matrix, index_to_activity, A):
    # Convert activity names in A to row indices
    A_indices = {i for i, activity in index_to_activity.items() if activity in A}

    # Sum over the selected rows to get total out-degree
    return sum(adj_matrix[i, :].sum() for i in A_indices)


def convert_activities_to_array(activity_set, mapping):
    x = np.array([mapping[activity] for activity in activity_set], dtype=np.int32)
    return x

def add_SE(nodes, st,en):
    if nodes & st:
        nodes.add('start')
    if nodes & en:
        nodes.add('end')
    return nodes

def out_degree_activity(adj_matrix, index_to_activity, activity):
    # Get the index of the activity
    activity_index = next((i for i, act in index_to_activity.items() if act == activity), None)

    # If the activity is not found, return 0
    if activity_index is None:
        return 0

    # Return the sum of the row corresponding to the activity (out-degree)
    return adj_matrix[activity_index, :].sum()


@njit
def get_input_output_B_indices(adj_matrix, B_indices):
    n = adj_matrix.shape[0]

    # Create a boolean mask for nodes in B
    B_mask = np.zeros(n, dtype=np.uint8)
    for i in range(B_indices.shape[0]):
        B_mask[B_indices[i]] = 1  # Mark nodes in B

    # Lists to store results
    input_B_list = []
    output_B_list = []

    for node in range(n):
        if B_mask[node]:  # Only check nodes in B
            has_incoming = False
            has_outgoing = False

            for neighbor in range(n):
                if not B_mask[neighbor]:  # Node is outside B
                    if adj_matrix[neighbor, node] > 0:  # Incoming edge
                        has_incoming = True
                    if adj_matrix[node, neighbor] > 0:  # Outgoing edge
                        has_outgoing = True

                # If both incoming & outgoing edges are found, break early
                if has_incoming and has_outgoing:
                    break

            if has_incoming:
                input_B_list.append(node)
            if has_outgoing:
                output_B_list.append(node)

    return np.array(input_B_list, dtype=np.int32), np.array(output_B_list, dtype=np.int32)

def aggregate_dictionaries(deviating, missing):
    aggregated_dict = {}

    for d in deviating:
        for key, value in d.items():
            if key in aggregated_dict:
                aggregated_dict[key]['deviating'] += value
            else:
                aggregated_dict[key] = {'deviating':value,'missing':0}

    for d in missing:
        for key, value in d.items():
            if key in aggregated_dict:
                aggregated_dict[key]['missing'] += value
            else:
                aggregated_dict[key] = {'deviating':0,'missing':value}
    return aggregated_dict

def aggregate_dictionaries2(deviating, missing):
    aggregated_dict = {'deviating':0,'missing':0}

    for d in deviating:
        aggregated_dict['deviating'] += np.sum(d)


    for d in missing:
        aggregated_dict['missing'] += np.sum(d)

    return aggregated_dict

def generate_nx_graph_from_log(log,nt, dummy_nodes):
    G = nx.DiGraph()
    for trace in log:
        tr_art = ('start',) + trace + ('end',)
        for i in range(len(tr_art) - 1):
            if G.has_edge(tr_art[i], tr_art[i + 1]):
                G[tr_art[i]][tr_art[i + 1]]['weight'] += log[trace]
            else:
                G.add_edge(tr_art[i], tr_art[i + 1], weight=log[trace])

    st = {}
    en = {}
    # Filter outgoing edges from start_node based on weight
    outgoing_edges = list(G.edges('start', data=True))
    for u, v, data in outgoing_edges:
        if data['weight'] < nt:
            G.remove_edge(u, v)
            if not nx.has_path(G, 'start', v):
               G.add_edge(u, v, weight=data['weight'])
               st[v] = data['weight']
        else:
            st[v] = data['weight']

    # Filter incoming edges to end_node based on frequency
    incoming_edges = list(G.in_edges('end', data=True))
    for u, v, data in incoming_edges:
        if data['weight'] < nt:
            G.remove_edge(u, v)
            if not nx.has_path(G, u, 'end'):
               G.add_edge(u, v, weight=data['weight'])
               en[u] = data['weight']
        else:
            en[u] = data['weight']
    for n in dummy_nodes:
        if n not in G.nodes:
            G.add_node(n)
    return G




def generate_nx_indirect_graph_from_log(log, dummy_nodes):
    # print("**************")
    G = nx.DiGraph()
    for trace in log:
        # print(trace)
        for i in range(0,len(trace)):
            if trace[i] not in G.nodes:
                G.add_node(trace[i])
            visited = set()
            for j in range(i+1,len(trace)):
                if trace[j] not in visited:
                    visited.add(trace[j])
                    if G.has_edge(trace[i], trace[j]):
                        G[trace[i]][trace[j]]['weight'] += log[trace]
                    else:
                        G.add_edge(trace[i], trace[j], weight=log[trace])
    for n in dummy_nodes:
        if n not in G.nodes:
            G.add_node(n)

    # print(G.nodes)
    return G

def read_append_write_json(file_path, new_item):
    # Read the current content of the JSON file
    with open(file_path, 'r') as file:
        data = json.load(file)

    # Append the new item to the list
    data.append(new_item)

    # Write the updated list back to the JSON file
    with open(file_path, 'w') as file:
        json.dump(data, file, indent=4)




from prolysis.util.functions import sum_out_degree_numba, sum_out_degree_single, get_edge_weight_numba, n_edges_numba
from numba import njit
import numpy as np



# @njit
def cost_seq(adj, A_indices, B_indices, sup, flow,min_cost,flag):
    # Compute total out-degree sums
    sum_out_degrees_A = sum_out_degree_numba(adj, A_indices)
    sum_out_degrees_B = sum_out_degree_numba(adj, B_indices)
    sum_out_degrees_total = sum_out_degrees_A + sum_out_degrees_B

    # Preallocate fixed-size storage for efficiency
    dev = 0
    mis = 0

    # Compute c1 and c2
    for i, x_idx in enumerate(A_indices):
        for j, y_idx in enumerate(B_indices):
            dev += get_edge_weight_numba(adj, y_idx, x_idx)

            if flag and dev> min_cost:
                return min_cost, dev, mis, False

            out_degree_x = sum_out_degree_single(adj, x_idx)
            out_degree_y = sum_out_degree_single(adj, y_idx)

            if sum_out_degrees_total > 0:
                mis += max(0, out_degree_x * sup * (out_degree_y / sum_out_degrees_total) -
                               get_edge_weight_numba(flow, x_idx, y_idx))
                if flag and dev > min_cost:
                    return min_cost, dev, mis, False
    if flag and dev < min_cost:
        min_cost = dev+mis

    return min_cost, dev, mis, True


# @njit
def cost_exc(adj, A_indices, B_indices,min_cost, flag):
    # Preallocate fixed-size storage for efficiency
    dev = 0

    # Compute c1
    for i, x_idx in enumerate(A_indices):
        for j, y_idx in enumerate(B_indices):
            dev += get_edge_weight_numba(adj, x_idx, y_idx)
            dev += get_edge_weight_numba(adj, y_idx, x_idx)
            if flag and dev > min_cost:
                return min_cost, dev, False
    if flag and dev < min_cost:
        min_cost = dev

    return min_cost, dev, True



def cost_exc_tau(adj, sup,start_index, end_index):
    # c = {}
    # if 'start' in net.nodes():
    mis = max(0, sup * sum_out_degree_single(adj, start_index) - get_edge_weight_numba(adj, start_index, end_index))
    return mis



# @njit
def cost_par(adj, A_indices, B_indices, sup,min_cost, flag):
    # Compute total out-degree sums
    sum_out_degrees_A = sum_out_degree_numba(adj, A_indices)
    sum_out_degrees_B = sum_out_degree_numba(adj, B_indices)
    sum_out_degrees_total = sum_out_degrees_A + sum_out_degrees_B

    # Preallocate fixed-size storage for efficiency
    mis = 0

    # Compute c1 and c2
    for i, a_idx in enumerate(A_indices):
        out_degree_a = sum_out_degree_single(adj, a_idx)
        for j, b_idx in enumerate(B_indices):
            out_degree_b = sum_out_degree_single(adj, b_idx)

            if sum_out_degrees_total > 0:
                mis += max(0, (out_degree_a * sup * out_degree_b) / sum_out_degrees_total -
                               get_edge_weight_numba(adj, a_idx, b_idx))
                mis += max(0, (out_degree_b * sup * out_degree_a) / sum_out_degrees_total -
                               get_edge_weight_numba(adj, b_idx, a_idx))
                if flag and mis > min_cost:
                    return min_cost, mis, False
    if flag and mis < min_cost:
        min_cost = mis

    return min_cost, mis, True


# @njit
def cost_loop(adj, A_indices, B_indices, sup, start_A_indices, end_A_indices, input_B_indices, output_B_indices, start_index, end_index,min_cost,flag):
    M_P = max(n_edges_numba(adj, output_B_indices, start_A_indices),
              n_edges_numba(adj, end_A_indices, input_B_indices))

    dev_cost = 0
    mis_cost = 0

    # Compute c1, c2, c3
    for x_idx in B_indices:
        dev_cost += get_edge_weight_numba(adj, start_index, x_idx) + get_edge_weight_numba(adj, x_idx, end_index)
        if flag and dev_cost > min_cost:
            return min_cost, dev_cost, mis_cost, False

        for y_idx in A_indices:
            if y_idx not in end_A_indices:
                dev_cost += get_edge_weight_numba(adj, y_idx, x_idx)
                if flag and dev_cost > min_cost:
                    return min_cost, dev_cost, mis_cost, False

            if y_idx not in start_A_indices:
                dev_cost += get_edge_weight_numba(adj, x_idx, y_idx)

            if flag and dev_cost > min_cost:
                    return min_cost, dev_cost, mis_cost, False

    # Compute c4 (Avoid np.array inside loops)
    if n_edges_numba(adj, output_B_indices, start_A_indices) > 0:
        for a_idx in start_A_indices:
            for b_idx in output_B_indices:
                start_edges = n_edges_numba(adj, np.array([start_index]), np.array([a_idx]))
                total_start_edges = n_edges_numba(adj, np.array([start_index]), start_A_indices)
                output_edges = n_edges_numba(adj, np.array([b_idx]), start_A_indices)
                total_output_edges = n_edges_numba(adj, output_B_indices, start_A_indices)

                if total_start_edges > 0 and total_output_edges > 0:
                    mis_cost += max(0, M_P * sup * (start_edges / total_start_edges) *
                                    (output_edges / total_output_edges) -
                                    get_edge_weight_numba(adj, b_idx, a_idx))

                if flag and dev_cost + mis_cost > min_cost:
                    return min_cost, dev_cost, mis_cost, False
    else:
        mis_cost += M_P * sup

    # Compute c5 (Avoid np.array inside loops)
    if n_edges_numba(adj, end_A_indices, input_B_indices) > 0:
        for a_idx in end_A_indices:
            for b_idx in input_B_indices:
                end_edges = n_edges_numba(adj, np.array([a_idx]), np.array([end_index]))
                total_end_edges = n_edges_numba(adj, end_A_indices, np.array([end_index]))
                input_edges = n_edges_numba(adj, end_A_indices, np.array([b_idx]))
                total_input_edges = n_edges_numba(adj, end_A_indices, input_B_indices)

                if total_end_edges > 0 and total_input_edges > 0:
                    mis_cost += max(0, M_P * sup * (end_edges / total_end_edges) *
                                    (input_edges / total_input_edges) -
                                    get_edge_weight_numba(adj, a_idx, b_idx))

                if flag and dev_cost + mis_cost > min_cost:
                    return min_cost, dev_cost, mis_cost, False
    else:
        mis_cost += M_P * sup

    if flag and dev_cost + mis_cost < min_cost:
        min_cost = dev_cost + mis_cost
    return min_cost,dev_cost, mis_cost, True




def cost_loop_tau(adj, sup, start_A_indices, end_A_indices,start_index, end_index):
    M_P = n_edges_numba(adj, end_A_indices, start_A_indices)


    start_sum = n_edges_numba(adj, np.array([start_index]), start_A_indices)
    end_sum = n_edges_numba(adj, end_A_indices, np.array([end_index]))
    mis_cost = 0

    for a_idx in start_A_indices:
        for b_idx in end_A_indices:
            if adj[b_idx,a_idx]>0:
                mis_cost += max(0, M_P * sup * (get_edge_weight_numba(adj,start_index,a_idx)/ start_sum)*
                                (get_edge_weight_numba(adj, b_idx, end_index)/ end_sum) -
                                get_edge_weight_numba(adj, b_idx, a_idx))
            else:
                mis_cost +=  M_P * sup * (get_edge_weight_numba(adj, start_index, a_idx) / start_sum) * (get_edge_weight_numba(adj, b_idx, end_index) / end_sum)
    return mis_cost


def overal_cost(costP,costM,ratio, size_par):
    ov_c= costP - ratio * size_par * costM
    return ov_c


import matplotlib
matplotlib.use('agg')
import pandas as pd
import pm4py
import shutil
import os
from prolysis.calls.minerful_calls import mine_minerful_for_declare_constraints, prune_constraints_minerful
import ruptures as rpt
import csv
from sklearn.metrics import mean_squared_error
from collections import defaultdict
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.cluster.hierarchy import fcluster
from scipy.spatial import KDTree
import json
from prolysis.util.redis_connection import redis_client
from prolysis.util.logging import log_command

linkage_method = 'ward'
linkage_metric = 'euclidean'

def export_constraints_per_cluster(constraints, constraints_json_path):
    dict_out = {}
    dict_out["constraints"] = []
    for constraint in constraints:
        temp_dict = {}
        temp_dict["template"] = constraint[0]
        temp_dict['parameters'] = []
        for par in constraint[1:3]:
            if par != '':
                temp_dict['parameters'].append([par.lstrip()])

        dict_out["constraints"].append(temp_dict)

    for key in range(len(dict_out['constraints'])):
        dict_out['constraints'][key]['template'] = dict_out['constraints'][key]['template'].rstrip().lstrip()

        dict_out['constraints'][key]['support'] = 0.99
        dict_out['constraints'][key]['confidence'] = 0.99

        for el in dict_out['constraints'][key]['parameters']:
            el[0] = el[0]

    with open(constraints_json_path, 'w') as fp:
        json.dump(dict_out, fp)
        fp.close()

def export_constraints_all_cluster(data_str,constraints_json_path,N_segment):
    for i in range(1,N_segment+1):
        dict_out = {}
        dict_out["tasks"] = json.loads(redis_client.get("tasks"))
        dict_out["constraints"] = []
        for cl in data_str.keys():
            for constraint in data_str[cl]:
                temp_dict = {}
                temp_dict["template"] = constraint['constraint type']
                temp_dict['parameters'] = []
                temp_dict['parameters'].append([constraint['first parameter'].lstrip()])
                if constraint['second parameter']!='':
                    temp_dict['parameters'].append([constraint['second parameter'].lstrip()])
                temp_dict['desirability'] = constraint[f'delta_segment_{i}']
                temp_dict['support'] = 0.0
                temp_dict['confidence'] = 0.0
                dict_out["constraints"].append(temp_dict)

        with open(constraints_json_path+f"all_{i}.json", 'w') as fp:
            json.dump(dict_out, fp,indent=4)
            fp.close()


def generate_features(w,kpi,n_bin):
    case_table = pd.read_csv("output_files/out.csv").sort_values(by=[kpi])
    ordered_case_ids = case_table['case_id']
    ordered_case_ids = ordered_case_ids.to_list()
    bin_size = round(len(case_table) / n_bin)
    event_table = pd.read_csv("output_files/out_event.csv")
    redis_client.set('tasks', json.dumps(list(set(event_table['activity_name']))))
    event_table['case:concept:name'] = event_table['case_id'].astype(str)
    event_table['concept:name'] = event_table['activity_name'].astype(str)
    event_table['time:timestamp'] = pd.to_datetime(event_table['timestamp'],format='mixed')
    event_table = event_table.sort_values('time:timestamp')
    event_table['case_id'] = pd.Categorical(event_table['case_id'], categories=ordered_case_ids, ordered=True)
    event_table = event_table.sort_values(by=['case_id', 'time:timestamp'], ascending=[True, True])
    event_table = event_table[['case:concept:name','concept:name','time:timestamp']]
    log_xes = pm4py.convert_to_event_log(event_table)
    pm4py.write_xes(log_xes, f"output_files/log_ordered.xes")

    window_size = 2 * w * bin_size
    sliding_window_size = bin_size
    mine_minerful_for_declare_constraints(window_size,sliding_window_size)

def find_all_supersets(graph,node, visited):
    for neighbor in graph[node]:
        if neighbor not in visited:
            visited.add(neighbor)
            find_all_supersets(graph,neighbor, visited)
    return visited

def correlation_calc(peaks,w,constraints,clusters_dict):
    n = len(peaks)
    peakmodif = [p-(w-1) for p in peaks]
    segments_sig = {}

    for i in range(n + 1):
        l = [0] * len(constraints[0])

        if i == 0:
            l[0:peakmodif[0]] = [1] * (peakmodif[0] - 0)

        elif i == n:
            l[peakmodif[n - 1]:] = [1] * (len(constraints[0]) - peakmodif[n - 1])

        else:
            l[peakmodif[i - 1]:peakmodif[i]] = [1] * (peakmodif[i] - peakmodif[i - 1])

        segments_sig[f'segment_{i}']=l




    corr_mat = []
    for seg in segments_sig.keys():
        target_array = np.array(segments_sig[seg])
        average_correlations = []

        for cluster in sorted([i for i in clusters_dict.keys()]):
            correlations = []
            for series in clusters_dict[cluster]:
                # Convert the current series to a NumPy array
                series_array = np.array(series[3:])

                # Calculate the correlation coefficient between target_series and this series
                correlation = np.corrcoef(target_array, series_array)[0, 1]

                # Append the correlation to the list
                correlations.append(correlation)
            average_correlation = np.mean(correlations)
            average_correlations.append(average_correlation)
        corr_mat.append(average_correlations)
    return corr_mat

def sort_by_closest_neigbour_HEADER(data):

    print('There were: ' + str(len(data)) + " values")

    # Convert data to numpy array for efficient operations
    data = np.array(data, dtype=object)

    # Initialize sorted data with the starting point (first point) and track remaining indices
    new_data = [data[0].tolist()]  # Start with the first point
    index_set = set(range(1, len(data)))  # Skip the first index, it's already in new_data

    # Track the original indices of points in the KDTree
    kd_tree_indices = list(index_set)
    kd_tree_data = data[kd_tree_indices]
    kd_tree = KDTree(kd_tree_data[:, 3:])

    while index_set:
        # Find nearest point in KDTree to the last point in `new_data`
        last_point = new_data[-1][3:]
        _, nearest_in_kd = kd_tree.query(last_point)

        # Map the KDTree index to the original data index
        min_ind = kd_tree_indices[nearest_in_kd]
        new_data.append(data[min_ind].tolist())


        # Update index_set and KDTree
        index_set.remove(min_ind)

        # Rebuild kd_tree for remaining points
        kd_tree_indices = list(index_set)
        kd_tree_data = data[kd_tree_indices]
        kd_tree = KDTree(kd_tree_data[:, 3:])

    return new_data

def import_minerful_constraints_timeseries_data(minerful_constraints_path,constraint_type_used):
    csvfile = open(minerful_constraints_path, 'r')
    csv_reader = csv.reader(csvfile, delimiter=';', quotechar='|')

    hea = next(csv_reader, None)
    hea2 = next(csv_reader, None)

    hea2 = hea2[2:]
    hea = hea[2:]

    header_output = list()

    for i in range(len(hea)):
        if i % 6 == 0:
            tem_h = [hea2[i][1:-1]]
            temp = hea[i]
            if temp[0] == '\'':
                temp = temp[1:]
            if temp[-1] == '\'':
                temp = temp[:-1]
            if temp[-1] == ')':
                temp = temp[:-1]
            # now we split the string
            name_of_constraint_end_index = temp.find('(')
            tem_h.append(temp[:name_of_constraint_end_index])
            temp = temp[name_of_constraint_end_index+1:]
            #find if we have two events or one
            separated_constraints_index = temp.find(', ')
            if not separated_constraints_index == -1:
                tem_h.append(temp[:separated_constraints_index])
                tem_h.append(temp[separated_constraints_index+1:])
            else:
                tem_h.append(temp)
                tem_h.append('')
        else:
            tem_h = [hea2[i][1:-1]] + tem_h[1:]

        header_output.append(tem_h)

    sequences = list()

    for i in range(len(hea)):
        sequences.append(list())

    corresponding_number_of_traces = []
    n_lines =0
    for r in csv_reader:
        corresponding_number_of_traces.append(r[:2])
        n_lines += 1
        counter = 0
        for i in range(len(r)):
            if counter > 1:
                sequences[i-2].append(float(r[i]))
            else:
                counter += 1


    constraints = []
    for i, j in zip(sequences, header_output):
        if j[0] == constraint_type_used:
            constraints.append(j[1:] + i)

    return constraints


def group_signals_by_type_activity(signals):
    # Create a nested defaultdict structure to handle the required dictionary format
    grouped_signals = defaultdict(lambda: defaultdict(set))

    for signal in signals:
        # Extract components
        signal_type = signal[0]  # The type of the signal
        activities = (signal[1], signal[2])  # Tuple of (activity1, activity2)
        signal_samples = tuple(signal[3:])  # Convert samples from position 3 onward to a tuple

        # Populate the dictionary
        if signal_type in {'RespondedExistence'}:
            grouped_signals[signal_samples][(activities[1].lstrip(), f' {activities[0]}')].add(
                'RespondedExistence_r')
            grouped_signals[signal_samples][activities].add(signal_type)
        else:
            grouped_signals[signal_samples][activities].add(signal_type)

    return grouped_signals

def prune_signals(theta_cvg):
    constraints_conf = import_minerful_constraints_timeseries_data(r"output_files/behavioral_signals.csv", 'Confidence')
    constraints_cov = import_minerful_constraints_timeseries_data(r"output_files/behavioral_signals.csv", 'Coverage')
    s_thr = theta_cvg * 100
    main_dim = constraints_conf
    filt_dim = constraints_cov

    not_include_templates = {'NotRespondedExistence',
                             'NotCoExistence',
                             'NotPrecedence',
                             'NotChainPrecedence',
                             'NotResponse',
                             'NotChainResponse',
                             'NotSuccession',
                             'NotChainSuccession'}

    filtered = []
    filtered_colored = []
    for i, c in enumerate(main_dim):
        filtered_sig = []
        filtered_colored_sig = []
        for j, v in enumerate(c):
            if j >= 3:
                if filt_dim[i][j] >= s_thr:
                    filtered_sig.append(v)
                    filtered_colored_sig.append(False)
                else:
                    filtered_sig.append(0)
                    if filt_dim[i][j] > 0:
                        filtered_colored_sig.append(True)
                    else:
                        filtered_colored_sig.append(False)
            else:
                filtered_sig.append(v)
                filtered_colored_sig.append(v)
        filtered.append(filtered_sig)
        filtered_colored.append(filtered_colored_sig)


    data = []
    data_color = []
    not_include = 0
    zeros_removed = 0
    hundreds_removed = 0
    nonchaning_removed = 0


    def all_list_in_interval(L):
        for i in L[1:]:
            if i < 1.05 * L[0] and i > L[0] * 0.95:
                continue
            else:
                return False
        return True

    for ind, j in enumerate(filtered):
        i = j[3:]
        if j[0] in not_include_templates:
            not_include += 1
        elif (mean_squared_error(i, [0] * len(i)) < 10):
            zeros_removed += 1
        elif mean_squared_error(i, [100] * len(i)) < 1:
            # print(j[0:3])
            hundreds_removed += 1
        elif all_list_in_interval(i):
            nonchaning_removed += 1
        else:
            data.append(j)
            data_color.append(filtered_colored[ind])

    print('there are : ' + str(len(data)) + " values left after deleting 100, and 0s")
    print("there were: " + str(not_include) + " vectors not included")
    print("there were: " + str(zeros_removed) + " vectors with zeros removed")
    print("there were: " + str(hundreds_removed) + " vectors with hundreds removed")
    print("there were: " + str(nonchaning_removed) + " vectors with non changing values removed")

    data_uncut = data

    ddd = group_signals_by_type_activity(data_uncut)
    subset_relations = [('RespondedExistence', 'Response'),
                        ('Response', 'AlternateResponse'),
                        ('AlternateResponse', 'ChainResponse'),
                        ('RespondedExistence', 'CoExistence'),
                        ('Response', 'Succession'),
                        ('AlternateResponse', 'AlternateSuccession'),
                        ('ChainResponse', 'ChainSuccession'),
                        ('CoExistence', 'Succession'),
                        ('Succession', 'AlternateSuccession'),
                        ('AlternateSuccession', 'ChainSuccession'),
                        ('RespondedExistence_r', 'Precedence'),
                        ('Precedence', 'AlternatePrecedence'),
                        ('AlternatePrecedence', 'ChainPrecedence'),
                        ('RespondedExistence_r', 'CoExistence'),
                        ('Precedence', 'Succession'),
                        ('AlternatePrecedence', 'AlternateSuccession'),
                        ('ChainPrecedence', 'ChainSuccession'),
                        ('AtLeast1', 'AtLeast2'),
                        ('AtLeast2', 'AtLeast3'),
                        ('AtMost3', 'AtMost2'),
                        ('AtMost2', 'AtMost1'),
                        ('AtMost1', 'Absence'),
                        ('AtLeast1', 'Init'),
                        ('AtLeast1', 'End')]
    subset_relations = [(a[1], a[0]) for a in subset_relations]

    templates = {'RespondedExistence',
                 'RespondedExistence_r',
                 'CoExistence',
                 'Precedence',
                 'AlternatePrecedence',
                 'ChainPrecedence',
                 'Response',
                 'AlternateResponse',
                 'ChainResponse',
                 'Succession',
                 'AlternateSuccession',
                 'ChainSuccession',
                 'NotRespondedExistence',
                 'NotCoExistence',
                 'NotPrecedence',
                 'NotChainPrecedence',
                 'NotResponse',
                 'NotChainResponse',
                 'NotSuccession',
                 'NotChainSuccession',
                 'Absence',
                 'AtLeast1',
                 'AtLeast2',
                 'AtLeast3',
                 'AtMost1',
                 'AtMost2',
                 'AtMost3',
                 'End',
                 'Init'}

    # Create a dictionary to store direct subset relationships
    graph = {letter: set() for letter in templates}

    # Populate the graph based on subset_relations

    for subset, superset in subset_relations:
        graph[subset].add(superset)

    pruned_list = []
    co_exist_list = set()
    for x in ddd.keys():
        for y in ddd[x].keys():
            maximal_letters = remove_subsets(ddd[x][y], subset_relations, graph)
            for dcl in maximal_letters - {'RespondedExistence_r'}:
                if dcl == 'CoExistence':
                    if (y[1].lstrip(), f' {y[0]}') not in co_exist_list:
                        pruned_list.append([dcl, y[0], y[1]] + list(x))
                        co_exist_list.add((y[0], y[1]))
                else:
                    pruned_list.append([dcl, y[0], y[1]] + list(x))
    return pruned_list, data_color

def remove_subsets(letters, subset_relations, graph):
    # Find all letters that are subsets of other letters
    subsets_to_remove = set()
    for letter in letters:
        # Find all supersets of the current letter
        supersets = find_all_supersets(graph, letter, set())
        subsets_to_remove.update(supersets)

    # The final set of letters excluding any that are subsets of another
    maximal_letters = letters - subsets_to_remove

    return maximal_letters

def clustering(pruned_list, linkage_method, linkage_metric, best_n_clusters):

    data_cut = []
    for data_point in pruned_list:
        data_cut.append(data_point[3:])

    '''build the clustering method'''
    Z = linkage(data_cut, method=linkage_method, metric=linkage_metric)  # metric='correlation'


    a = fcluster(Z, best_n_clusters, 'maxclust')
    clusters_dict = dict()
    for cluster_n, data in zip(a, pruned_list):
        if not cluster_n in clusters_dict:
            clusters_dict[cluster_n] = [data]
        else:
            clusters_dict[cluster_n].append(data)

    constraints = []
    cluster_bounds = []
    count = 0
    clusters_with_declare_names = {}

    order_cluster = [(i, len(clusters_dict[i])) for i in clusters_dict.keys()]
    # order_cluster = sorted(order_cluster, key=lambda x : -x[1])
    order_cluster = sorted(order_cluster, key=lambda x: -x[0])
    order_cluster = [key for key, _ in order_cluster]

    for key in order_cluster:
        # preprocess the clusters for plotting better (sorting them by similarity)
        clusters_dict[key] = sort_by_closest_neigbour_HEADER(clusters_dict[key])

        clusters_with_declare_names[key] = [clusters_dict[key][0][:3]]
        constraints.append(clusters_dict[key][0][3:])
        for i in clusters_dict[key][1:]:
            clusters_with_declare_names[key].append(i[:3])
            constraints.append(i[3:])
        count += len(clusters_dict[key])
        cluster_bounds.append(count)
        print("Cluster size: " + str(len(clusters_dict[key])))

    print('number of clusters: ' + str(len(clusters_dict)))
    print("clusters were ordered in the : ")
    print(order_cluster)

    return order_cluster, clusters_with_declare_names, cluster_bounds, clusters_dict, constraints
def constraints_export(clusters_with_declare_names, peaks, w,clusters_dict):
    file_path = r"output_files/"

    import statistics
    def generate_natural_language(constraint):
        c, a, b = constraint
        if c == "RespondedExistence":
            description = f"If {a} occurs, {b} occurs as well."
        elif c == "CoExistence":
            description = f"{a} and {b} always occur together."
        elif c == "Response":
            description = f" If {a} occurs, then {b} occurs after it."
        elif c == "AlternateResponse":
            description = f"If {a} occurs, then {b} occurs afterwards before {a} recurs."
        elif c == "ChainResponse":
            description = f"If {a} occurs, then {b} occurs immediately after it."
        elif c == "Precedence":
            description = f"{b} occurs only if preceded by {a}."
        elif c == "AlternatePrecedence":
            description = f"{b} occurs only if preceded by {a} with no other {b} in between."
        elif c == "ChainPrecedence":
            description = f"{b} occurs only if {a} occurs immediately before it. "
        elif c == "Succession":
            description = f"{a} occurs if and only if it is followed by {b}."
        elif c == "AlternateSuccession":
            description = f"{a} and {b} occur if and only if they follow one another, alternating."
        elif c == "ChainSuccession":
            description = f"{a} and {b} occurs if and only if {b} immediately follows {a}."
        elif c == "Init":
            description = f"{a} is the first to occur."
        elif c == "End":
            description = f"{a} is the last to occur."
        elif c == "Absence":
            description = f"{a} must never occur."
        elif c == "AtMost1":
            description = f"{a} occurs at most once."
        elif c == "AtMost2":
            description = f"{a} occurs at most two times."
        elif c == "AtMost3":
            description = f"{a} occurs at most three times."
        elif c == "AtLeast1":
            description = f"{a} occurs at least once."
        elif c == "AtLeast2":
            description = f"{a} occurs at least two times."
        elif c == "AtLeast3":
            description = f"{a} occurs at least three times."
        else:
            description = f"The constraint is un known!"

        return description

    def stat_extract(value, peaks, w):
        new_list = []
        for const in value:
            stat_dic = {'constraint type': const[0], 'first parameter': const[1], 'second parameter': const[2],
                        'description': generate_natural_language(const[0:3])}
            c = const[3:]
            seg_num = 1
            for p in range(len(peaks) + 1):
                if seg_num == 1:
                    stat_dic[f'segment_{seg_num}'] = round(statistics.mean(c[0:(peaks[p] - (w - 1) + 1)]), 2)
                    stat_dic[f'~segment_{seg_num}'] = round(statistics.mean(c[(peaks[p] - (w - 1)):]), 2)
                elif seg_num <= len(peaks):
                    stat_dic[f'segment_{seg_num}'] = round(
                        statistics.mean(c[(peaks[p - 1] - (w - 1)):(peaks[p] - (w - 1) + 1)]), 2)
                    stat_dic[f'~segment_{seg_num}'] = round(
                        statistics.mean(c[0:(peaks[p - 1] - (w - 1) + 1)] + c[(peaks[p] - (w - 1)):]), 2)
                else:
                    stat_dic[f'segment_{seg_num}'] = round(statistics.mean(c[(peaks[p - 1] - (w - 1)):]), 2)
                    stat_dic[f'~segment_{seg_num}'] = round(statistics.mean(c[0:(peaks[p - 1] - (w - 1) + 1)]), 2)
                stat_dic[f'delta_segment_{seg_num}'] = round(
                    stat_dic[f'segment_{seg_num}'] - stat_dic[f'~segment_{seg_num}'], 2)
                seg_num += 1
            new_list.append(stat_dic)
        return new_list


    data_str_keys = {str(key): stat_extract(value, peaks, w) for key, value in clusters_dict.items()}
    # Write the dictionary to a JSON file
    with open(file_path+"data1.json", 'w') as file:
        json.dump(data_str_keys, file, indent=4)

    export_constraints_all_cluster(data_str_keys,file_path,len(peaks)+1)
    for cl in clusters_with_declare_names.keys():
        export_constraints_per_cluster(clusters_with_declare_names[cl], file_path + f'constraints_{cl}.json')
        # prune_constraints_minerful(file_path + f'constraints_{cl}.json', file_path + f'constraints_{cl}_pruned.csv')




    print(f"Dictionary successfully saved to {file_path}")

def PELT_change_points(order_cluster,clusters_dict):
    def dingOptimalNumberOfPoints(algo):
        point_detection_penalty = 50
        x_lines = algo.predict(pen=point_detection_penalty)

        while point_detection_penalty >= len(x_lines):
            point_detection_penalty -= 1
            x_lines = algo.predict(pen=point_detection_penalty)

        if len(x_lines) > 15:
            x_lines = x_lines[-1:]
        return x_lines

    horisontal_separation_bounds_by_cluster = {}
    # in ths case we want to detect the drifts in the whole range of constrains at the same time

    dd = []
    for dk in order_cluster:
        for i in clusters_dict[dk]:
            dd.append(i[3:])

    sig = np.array(dd)
    signal = np.transpose(sig)
    # algo = rpt.Pelt(model="rbf", custom_cost=c).fit(signal)
    algo = rpt.Pelt(model="rbf").fit(signal)
    x_lines = dingOptimalNumberOfPoints(algo)
    horisontal_separation_bounds_by_cluster[0] = x_lines
    # pen - penalizing the number of change points

    return x_lines


import plotly.graph_objects as go
import numpy as np


def plot_figures(df, masks, n_bin, map_range, peaks, constraints, w, cluster_bounds, clusters_with_declare_names,
                 data_color, corr_mat, WINDOWS):
    # every = 1
    color_theme_drift_map = 'Blues'


    ################## Figure 3: Control Flow Features and Change Points ####################
    L1 = []
    for k in clusters_with_declare_names.keys():
        for s in clusters_with_declare_names[k]:
            L1.append(s)
    L1_index = {tuple(sublist[0:3]): idx for idx, sublist in enumerate(L1)}

    # Sort and filter L2 based on L1
    L2_ordered = sorted(
        [sublist for sublist in data_color if tuple(sublist[0:3]) in L1_index],
        key=lambda x: L1_index[tuple(x[0:3])]
    )
    L2_ordered = [x[3:] for x in L2_ordered]

    # Prepare data for the heatmap
    data_c = []
    for i in range(len(constraints)):
        new_list = [np.nan] * (w - 1) + constraints[i] + [np.nan] * (w - 1)
        data_c.append(new_list)

    # Create the first subplot (Heatmap with clusters and peaks)
    fig3 = go.Figure()
    ordered_keys = sorted(L1_index, key=L1_index.get)
    ordered_keys =[str(x) for x in ordered_keys]
    # Add heatmap
    fig3.add_trace(go.Heatmap(
        z=data_c,
        y = ordered_keys,
        colorscale=color_theme_drift_map,
        zmin=0,
        zmax=100,
        colorbar=dict(
            title=dict(
                text = "Confidence",
            font=dict(size=18)),
            tickfont=dict(size=16)
        )
    ))

    # Add cluster boundary lines
    for bound in cluster_bounds:
        fig3.add_shape(
            type="line",
            x0=0,
            x1=n_bin-2,
            y0=bound-0.5,
            y1=bound-0.5,
            line=dict(color="black", dash="dash", width=2)
        )

    # Add vertical peak lines
    for pp in peaks:
        fig3.add_shape(
            type="line",
            x0=pp,
            x1=pp,
            y0=0,
            y1=len(data_c),
            line=dict(color="red", dash="dash", width=4)
        )

    # Set x-axis ticks and labels
    x_tick = []
    x_labels = []
    for i in range(len(peaks) + 1):
        if i == 0:
            x_tick.append(round(peaks[i] / 2, 0))
        elif i == len(peaks):
            x_tick.append(round((n_bin + peaks[i - 1]) / 2, 0))
        else:
            x_tick.append(round((peaks[i] + peaks[i - 1]) / 2, 0))
        x_labels.append(f'segment {i + 1}')

    fig3.update_layout(
        plot_bgcolor="gray",  # Gray plot background
        title="Control Flow Features and Change Points Overview",
        xaxis=dict(
            title="Segments",
            tickvals=x_tick,
            ticktext=x_labels,
            tickfont=dict(size=16)
        ),
        yaxis=dict(
            title="Clusters",
            tickvals=[round((cb + cluster_bounds[i - 1]) / 2, 0) if i > 0 else round(cb / 2, 0) for i, cb in
                      enumerate(cluster_bounds)],
            ticktext=[f'cluster {len(cluster_bounds) - i}' for i in range(len(cluster_bounds))],
            tickfont=dict(size=16)
        ),
        template="plotly_white",
        height=500,
        width=750
    )

    ##################### Figure 4: Correlation Heatmap #####################
    corr_mat_transposed = np.array(corr_mat).T  # Transpose the matrix

    # Create correlation heatmap
    fig4 = go.Figure()

    fig4.add_trace(go.Heatmap(
        z=corr_mat_transposed,
        colorscale=[[0.0, '#8B0000'], [0.5, 'white'], [1.0, '#00008B']],
        zmin=-1,
        zmax=1,
        colorbar=dict(
            title=dict(
                text = "Correlation",
            font=dict(size=18)),
            tickfont=dict(size=16)
        )
    ))

    fig4.update_layout(
        title="Correlation Between Features and Segments",
        xaxis=dict(
            title="Segments",
            tickvals=np.arange(len(x_labels)),
            ticktext=x_labels,
            tickfont=dict(size=16)
        ),
        yaxis=dict(
            title="Features",
            tickvals=np.arange(len(cluster_bounds)),
            ticktext=[f'cluster {i+1}' for i in range(len(cluster_bounds))],
            tickfont=dict(size=16)
        ),
        template="plotly_white",
        height=600,
        width=600
    )

    return fig3, fig4


# We made the visualization in Plotly later, to allow for intractive visualization
# def plot_figures(df, masks, n_bin, map_range, peaks, constraints, w, cluster_bounds,clusters_with_declare_names, data_color, corr_mat,WINDOWS):
#     every = 2
#     color_theme_drift_map = 'Blues'
#
#     def insert_and_clean_np(array, new_number):
#         """
#         Insert a new number into a sorted NumPy array, maintain the order,
#         and remove any numbers within a distance of 1 from the new number.
#
#         Parameters:
#         - array (np.ndarray): A sorted NumPy array of numbers.
#         - new_number (int or float): The number to insert.
#
#         Returns:
#         - np.ndarray: The updated NumPy array.
#         """
#         if new_number not in array:
#             # Find the correct insertion point
#             insertion_index = np.searchsorted(array, new_number)
#
#             # Insert the new number into the array
#             array = np.insert(array, insertion_index, new_number)
#
#             # Identify elements to keep (distance > 1 from the new number)
#             mask = np.abs(array - new_number) != 1
#
#             # Return the updated array
#             return array[mask]
#             # return mask
#         else:
#             return array
#
#     ################## Figure 3############################################
#     L1 = []
#     for k in clusters_with_declare_names.keys():
#         for s in clusters_with_declare_names[k]:
#             L1.append(s)
#     L1_index = {tuple(sublist[0:3]): idx for idx, sublist in enumerate(L1)}
#
#     # Sort and filter L2 based on L1
#     L2_ordered = sorted([sublist for sublist in data_color if tuple(sublist[0:3]) in L1_index],
#                         key=lambda x: L1_index[tuple(x[0:3])])
#     L2_ordered = [x[3:] for x in L2_ordered]
#     data_c = []
#     data_c_color_1 = []
#     data_c_color_2 = []
#     for i in range(len(constraints)):
#         new_list = [0] * (w - 1) + constraints[i] + [0] * (w - 1)
#         new_list_color_1 = [True] * (w - 1) + L2_ordered[i] + [True] * (w - 1)
#         new_list_color_2 = [True] * (w - 1) + [False] * len(L2_ordered[i]) + [True] * (w - 1)
#         data_c.append(new_list)
#         data_c_color_1.append(new_list_color_1)
#         data_c_color_2.append(new_list_color_2)
#
#     # Create a new figure with two subplots with different heights
#     fig3, (ax1, ax2) = plt.subplots(2, 1, figsize=(15, 9),
#                                    gridspec_kw={'height_ratios': [10, 1]})  # First plot 3 times bigger
#
#     fig3.suptitle('Control Flow Features and Change Points Overview', fontsize=20)
#
#     # sns.heatmap(data_c, mask=list(mask_ax1[windows.index(w)]), linewidth=0, cmap=color_theme_drift_map,  vmin=0, vmax=100, ax=ax1)
#     light_gray_cmap = ListedColormap(['#d3d3d3'])
#     sns.heatmap(data_c_color_1, mask=np.array(data_c_color_2), cmap=light_gray_cmap, cbar=False, ax=ax1)
#     ax1.set_facecolor("gray")
#     sns.heatmap(np.array(data_c), mask=np.array(data_c_color_1), linewidth=0, cmap=color_theme_drift_map, cbar=True,
#                 vmin=0, vmax=100, ax=ax1)
#
#     cbar1 = ax1.collections[1].colorbar
#     cbar1.ax.tick_params(labelsize=16)
#     cbar1.set_label('Confidence', fontsize=18)
#
#     original_ticks = cbar1.get_ticks()  # Get the current ticks (e.g., [0, 20, 40, ..., 100])
#     normalized_ticks = np.linspace(0, 1, len(original_ticks))  # Map the ticks to 0–1
#
#     # Update the color bar with normalized ticks
#     cbar1.set_ticks(original_ticks)  # Retain original tick positions
#     cbar1.set_ticklabels([f"{tick:.1f}" for tick in normalized_ticks])  # Set new labels (0–1)
#
#     for bound in cluster_bounds:
#         ax1.axhline(y=bound, color='black', linestyle='--', linewidth=2)
#
#     for pp in peaks:
#         ax1.axvline(x=0.5 + pp - 0.05, color='red', linestyle='--', linewidth=4)
#
#
#     ax1.set_xticks([])
#     y_tick = []
#     y_labels = []
#     for i, cb in enumerate(cluster_bounds):
#         if i == 0:
#             y_tick.append(round(cb / 2, 0))
#         else:
#             y_tick.append(round((cluster_bounds[i] + cluster_bounds[i - 1]) / 2, 0))
#         y_labels.append(f'cluster {len(cluster_bounds) - i}')
#     ax1.set_yticks(y_tick)
#     ax1.set_yticklabels(y_labels)
#
#     x_tick = []
#     x_labels = []
#     for i in range(len(peaks) + 1):
#         if i == 0:
#             x_tick.append(round(peaks[i] / 2, 0))
#         elif i == len(peaks):
#             x_tick.append(round((n_bin + peaks[i - 1]) / 2, 0))
#         else:
#             x_tick.append(round((peaks[i] + peaks[i - 1]) / 2, 0))
#         x_labels.append(f'segment {i + 1}')
#     ax1.set_xticks(x_tick)
#     ax1.set_xticklabels(x_labels)
#     ax1.tick_params(axis='x', labelsize=16, rotation=0)
#     ax1.tick_params(axis='y', labelsize=16, rotation=0)
#     ax1.grid(False)
#
#
#     sns.heatmap(df.loc[w].to_frame().T, mask=masks[WINDOWS.index(w)], cmap="Reds", linewidth=0,
#                 ax=ax2)
#
#     ticks = np.arange(0, n_bin - 1, 2)
#     pk_id = []
#     for new_number in peaks:
#         ticks = insert_and_clean_np(ticks, new_number)
#         pk_id.append(np.where(ticks == new_number)[0][0])
#     ticks_labels = [str(round(x * (100 / n_bin))) + "% (" + str(round(map_range[str(x)], 1)) + ")" for x in (ticks + 1)]
#
#     print(pk_id)
#     ax2.set_xticks(0.5 + ticks)
#     ax2.set_xticklabels(ticks_labels)
#
#     tick_labels = ax2.get_xticklabels()  # Get all x-axis tick labels
#     print(tick_labels)
#     for pk in pk_id:
#         # Change the color of specific ticks
#         tick_labels[pk].set_color('red')  # Change tick at position 2 to red
#
#     ax2.set_facecolor("gray")
#     ax2.grid(False)
#     # ax2.set_title('sliding window analysis', fontsize=20)
#     ax2.set_xlabel('traces', fontsize=18)
#     ax2.set_ylabel('window size', fontsize=18)
#     ax2.tick_params(axis='x', labelsize=16, rotation=90)
#     ax2.tick_params(axis='y', labelsize=16)
#
#     # Adjust colorbar font size
#     cbar2 = ax2.collections[0].colorbar
#     cbar2.ax.tick_params(labelsize=16)
#     cbar2.set_label('ldist', fontsize=18)
#
#     ##################### Figure 4####################
#     from matplotlib.colors import LinearSegmentedColormap
#
#     corr_mat_transposed = np.array(corr_mat).T  # Transpose the matrix
#
#     # Output the list of correlations
#     fig4 = plt.figure(figsize=(8, 7))
#     fig4.suptitle('Correlation Between Features and Segments', fontsize=20)
#
#     colors = ['#8B0000', 'white', '#00008B']  # Dark blue, white, dark red
#     # Create a colormap
#     custom_cmap = LinearSegmentedColormap.from_list("custom_cmp", colors)
#
#     # Create a heatmap with black grid lines
#     ax3 = sns.heatmap(
#         corr_mat_transposed,
#         annot=False,
#         cmap=custom_cmap,
#         cbar=True,
#         vmin=-1,
#         vmax=1,
#         linewidths=0.5,
#         linecolor='gray'
#     )
#
#     ax3.set_xticks(0.5 + np.arange(0, len(peaks) + 1))
#     ax3.set_xticklabels(x_labels)
#
#     ax3.set_yticks(0.5 + np.arange(0, len(cluster_bounds)))
#     ax3.set_yticklabels(y_labels[::-1])
#
#     # Rotate the y-axis tick labels instead of x-axis
#     ax3.tick_params(axis='y', rotation=0, labelsize=16)
#     ax3.tick_params(axis='x', labelsize=16)
#
#     ax3.invert_yaxis()
#
#     cbar3 = ax3.collections[0].colorbar
#     cbar3.ax.tick_params(labelsize=16)
#     cbar3.set_label('correlation', fontsize=18)
#
#     buf = BytesIO()
#     fig3.savefig(buf, format="png", bbox_inches='tight')
#     fig_data3 = base64.b64encode(buf.getbuffer()).decode("ascii")
#
#     buf = BytesIO()
#     fig4.savefig(buf, format="png", bbox_inches='tight')
#     fig_data4 = base64.b64encode(buf.getbuffer()).decode("ascii")
#
#     return f'data:image/png;base64,{fig_data3}', f'data:image/png;base64,{fig_data4}'



def report(data, cluster, segment):
    if f'segment_{segment}' not in data[str(cluster)][0].keys():
        print('segment does not exist')
    else:
        file_path = r"output_files/"
        prune_constraints_minerful(file_path + f'constraints_{cluster}.json', file_path + f'constraints_{cluster}_pruned.csv')
        fp = f"output_files/constraints_{cluster}_pruned.csv"
        df = pd.read_csv(fp, sep=";")
        df = df.applymap(lambda x: x.strip("'\"") if isinstance(x, str) else x)
        df.columns = df.columns.str.strip("'\"")
        df.set_index('Constraint', inplace=True)
        list = []
        for d in data[str(cluster)]:
            if d['second parameter'] == '':
                const = f"{d['constraint type']}({d['first parameter']})"
                if df.loc[f'{const}', 'Redudant'] == False:
                    list.append((d['description'], round(d[f'segment_{segment}'] - d[f'~segment_{segment}'], 2)))

            else:
                const = f"{d['constraint type']}({d['first parameter']},{d['second parameter']})"
                if df.loc[f'{const}', 'Redudant'] == False:
                    list.append((d['description'], round(d[f'segment_{segment}'] - d[f'~segment_{segment}'], 2)))

        list_sorted = sorted(list, key=lambda x: x[1], reverse=False)[0:]
        list_sorted_reverse = sorted(list, key=lambda x: x[1], reverse=True)[0:]

        return [f'rank {k+1}: {x[0]}, with score {x[1]}' for k,x in enumerate(list_sorted)], [f'rank {k+1}: {x[0]}, with score {x[1]}' for k,x in enumerate(list_sorted_reverse)]
def decl2NL(cluster, segment):
    file_path = r"output_files/data1.json"
    with open(file_path, 'r') as file:
        data = json.load(file)
    list_sorted, list_sorted_reverse = report(data, cluster, segment)
    return list_sorted, list_sorted_reverse



def apply_feature_extraction(theta_cvg):
    """Main function to apply the analysis."""
    pruned_list, data_color = prune_signals(theta_cvg)
    redis_client.set("pruned_list",json.dumps(pruned_list))
    redis_client.set("data_color",json.dumps(data_color))
    return "Feature Generation Done!"

def apply_X(n_bin, w, n_clusters):
    """Main function to apply the analysis."""
    # if w not in WINDOWS:
    #     WINDOWS.append(w)
    #     WINDOWS.sort(reverse=True)
    WINDOWS = [w]

    pruned_list = json.loads(redis_client.get("pruned_list"))
    data_color = json.loads(redis_client.get("data_color"))

    log_command("clustering started!")
    order_cluster, clusters_with_declare_names, cluster_bounds, clusters_dict, constraints = clustering(pruned_list,
                                                                                                        linkage_method,
                                                                                                        linkage_metric,
                                                                                                        n_clusters)
    log_command("clustering done!")

    df = pd.read_json(redis_client.get("df"), orient="split")
    masks = json.loads(redis_client.get("masks"))
    map_range = json.loads(redis_client.get("map_range"))
    peaks = json.loads(redis_client.get("peaks"))


    redis_client.set("segments_count",len(peaks) + 1)
    redis_client.set("clusters_count", len(clusters_with_declare_names.keys()))

    # PELT_change_points(order_cluster, clusters_dict)
    log_command("exporting the constraints started!")
    constraints_export(clusters_with_declare_names, peaks, w, clusters_dict)
    log_command("exporting the constraints done!")

    log_command("correlation matrix is being calculated!")
    corr_mat = correlation_calc(peaks, w, constraints, clusters_dict)
    log_command("correlation matrix calculated!")

    log_command("figure 3 and 4 are being generated!")
    fig3_path, fig4_path = plot_figures(df, masks, n_bin, map_range, peaks,
                                        constraints, w, cluster_bounds,
                                        clusters_with_declare_names, data_color, corr_mat,
                                        WINDOWS)
    log_command("figure 3 and 4 are generated!")
    return fig3_path, fig4_path


from prolysis.util.functions import add_SE, select_submatrix, bfs_descendants, bfs_ancestors, get_edge_weight_numba, n_edges_numba, convert_activities_to_array
from prolysis.discovery.candidate_search.is_allowed_2 import is_allowed
import numpy as np


def find_possible_partitions(rules, st_net, en_net,fp, adj_matrix,nodes_order,mapping,adj_dict):
    # activity_list = set(net.nodes) - {'start', 'end'}
    
    activity_list = set(mapping.keys()) - {'start', 'end'}
    queue = [(set(), {'start'})]
    visited = set()
    reserve = []

    nodes_order_np = np.array([mapping[s] for s in nodes_order], dtype=np.int16)

    block = False

    lowest_cost = 10000

    if get_edge_weight_numba(adj_matrix, mapping['start'], mapping['end']) > 0:
        if len(rules[1]) > 0:
            na, block, penalty,lowest_cost = is_allowed(activity_list, set(), rules,{'exc_tau'},lowest_cost)
        else:
            na = set()
            penalty = {'seq': 0, 'exc': 0, 'par': 0, 'loop': 0, 'loop_tau': 0, 'exc_tau':0}
        possible_cuts = {'exc_tau'} - na
        reserve += [(activity_list, set(), ct, penalty[ct]) for ct in possible_cuts]


    if (n_edges_numba(adj_matrix, convert_activities_to_array(en_net,mapping), convert_activities_to_array(st_net,mapping)) > 0) and st_net==set([x for x in st_net if (x,x) in fp.edges() and fp[x][x]['weight'] > 0]):
        if len(rules[1]) > 0:
            na, block, penalty,lowest_cost = is_allowed(activity_list, set(), rules,{'loop_tau'},lowest_cost)
        else:
            na = set()
            penalty = {'seq': 0, 'exc': 0, 'par': 0, 'loop': 0, 'loop_tau': 0, 'exc_tau':0}
        possible_cuts = {'loop_tau'} - na
        reserve += [(activity_list, set(), ct, penalty[ct]) for ct in possible_cuts]


    while len(queue) != 0:
        current = queue.pop()
        current_set, current_adj = current

        for x in current_adj:
            new_state = current_set | {x}
            new_state = add_SE(new_state, st_net, en_net)

            if frozenset(new_state) not in visited:
                new_adj = (current_adj | adj_dict[x]) - new_state

                visited.add(frozenset(new_state))

                B = activity_list - new_state

                if (len(B) == 0) or (len(B) == len(activity_list)):
                    queue.append((new_state, new_adj))
                    continue

                B = add_SE(B,st_net,en_net)
                new_state_np = np.array([mapping[s] for s in new_state], dtype=np.int16)
                B_np = np.array([mapping[s] for s in B], dtype=np.int16)

                adj_A = select_submatrix(adj_matrix, nodes_order_np, new_state_np)
                adj_B = select_submatrix(adj_matrix, nodes_order_np, B_np)


                # 'start' ~> netB
                if 'start' in B:
                    descendants = bfs_descendants(adj_B, np.where(B_np==mapping['start'])[0][0])
                    start2B = (descendants.shape[0] == (B_np.shape[0]-1))
                else:
                    start2B = False
                # 'end' ~> netA
                if 'end' in new_state:
                    ancestors = bfs_ancestors(adj_A, np.where(new_state_np==mapping['end'])[0][0])
                    A2end = (ancestors.shape[0] == (new_state_np.shape[0]-1))
                else:
                    A2end = False

                # 'end' ~> netB
                if 'end' in B:
                    ancestors = bfs_ancestors(adj_B, np.where(B_np==mapping['end'])[0][0])
                    B2end = (ancestors.shape[0] == (B_np.shape[0] - 1))
                else:
                    B2end = False


                possible_cuts = set()
                if B2end:
                    possible_cuts.add("seq")
                    if A2end:
                        if n_edges_numba(adj_matrix, convert_activities_to_array(B-{'start','end'}, mapping), convert_activities_to_array(new_state & st_net, mapping)) != 0 and n_edges_numba(adj_matrix,convert_activities_to_array(new_state & en_net,mapping), convert_activities_to_array(B-{'start','end'},mapping)) != 0:
                                possible_cuts.add("loop")
                        if start2B and (B not in visited):
                            possible_cuts.update(["exc", "par"])
                elif A2end:
                    if n_edges_numba(adj_matrix, convert_activities_to_array(B,mapping), convert_activities_to_array(new_state & st_net,mapping)) != 0 and n_edges_numba(adj_matrix, convert_activities_to_array(new_state & en_net,mapping),convert_activities_to_array(B,mapping)) != 0:
                            possible_cuts.add("loop")

                if possible_cuts:
                    if len(rules[1]) > 0:
                        na, block, penalty,lowest_cost = is_allowed(new_state, B, rules,possible_cuts,lowest_cost)
                    else:
                        penalty = {'seq': 0, 'exc': 0, 'par': 0, 'loop': 0, 'loop_tau': 0, 'exc_tau': 0}
                    reserve += [(new_state, B, ct, penalty[ct]) for ct in possible_cuts]


                if not block:
                    queue.append((new_state, new_adj))
    # print(penalty)
    min_value = min(reserve, key=lambda x: x[3])[3]
    if min_value>0:
        print(f"warning! the minimum cost is {min_value} and not 0")
    result = [(tup[0],tup[1],tup[2]) for tup in reserve if tup[3] == min_value]
    # return valid,reserve
    return result,reserve


from collections import Counter

def split(cut_type, cut, l):
    LA = Counter()
    LB = Counter()


    if cut_type == 'seq':
        for tr in l:
            if tr:
                cost = [sum(ev in cut[1] for ev in tr[:i]) + sum(ev in cut[0] for ev in tr[i:]) for i in range(len(tr) + 1)]
                split_point = cost.index(min(cost))
                trace_A = tuple(x for x in tr[0:split_point] if x in cut[0])
                trace_B = tuple(x for x in tr[split_point:] if x in cut[1])
                LA[trace_A] += l[tr]
                LB[trace_B] += l[tr]
            # if not tr:
            #     LA[()] += l[tr]
            #     LB[()] += l[tr]


    elif cut_type == 'exc':
        for tr in l:
            if tr:
                A_count = sum(1 for ev in tr if ev in cut[0])
                B_count = len(tr) - A_count
                target = cut[0] if A_count >= B_count else cut[1]
                T = tuple(ev for ev in tr if ev in target)
                (LA if A_count >= B_count else LB)[T] += l[tr]
            # if not tr:
            #     LA[()] += l[tr]
            #     LB[()] += l[tr]

    elif cut_type == 'exc2' or cut_type == 'exc_tau':
        for tr in l:
            if not tr:
                LB[()] = l[tr]
                continue
            A_count = sum(1 for ev in tr if ev in cut[0])
            B_count = len(tr) - A_count
            target = cut[0] if A_count >= B_count else cut[1]
            T = tuple(ev for ev in tr if ev in target)
            (LA if A_count >= B_count else LB)[T] += l[tr]



    elif cut_type == 'par':
        for tr in l:
            if tr:
                T1 = tuple([ev for ev in tr if ev in cut[0]])
                T2 = tuple([ev for ev in tr if ev in cut[1]])
                LA[T1] += l[tr]
                LB[T2] += l[tr]
            # if not tr:
            #     LA[()] += l[tr]
            #     LB[()] += l[tr]

    elif cut_type == 'loop':
        for tr in l:
            if tr:
                flagA = tr[0] in cut[0]
                # if not flagA:
                #     LA[()] += l[tr]
                T = ()
                for ind, ev in enumerate(tr):
                    T = T + (ev,)
                    if ind < len(tr) - 1:
                        if flagA and tr[ind + 1] in cut[1]:
                            LA[T] += l[tr]
                            T = ()
                            flagA = False
                        elif not flagA and tr[ind + 1] in cut[0]:
                            LB[T] += l[tr]
                            T = ()
                            flagA = True
                    else:
                        if flagA:
                            LA[T] += l[tr]
                        else:
                            # LA[()] += l[tr]
                            LB[T] += l[tr]
            # if not tr:
            #     LA[()] += l[tr]


    elif cut_type == 'loop1':
        for tr in l:
            if tr:
                if len(tr) == 1:
                    LA[tr] += l[tr]
                else:
                    LA[(tr[0]),] += l[tr]
                    for _ in tr[1:]:
                        LB[()] += l[tr]
                        LA[(tr[0]),] += l[tr]
            # if not tr:
            #     LA[()] += l[tr]



    elif cut_type == 'loop_tau':
        # from collections import Counter
        st_acts = cut[0]
        en_acts = cut[1]
        for tr in l:
            if tr:
                T = ()
                for i, ev in enumerate(tr):
                    T = T + (ev,)
                    if i < len(tr) - 1 and tr[i] in en_acts and tr[i + 1] in st_acts:
                        # if max(Counter(T+(tr[i+1],)).values())>1:
                        # if tr[i] in en_acts and tr[i + 1] in st_acts:
                        LA[T] += l[tr]
                        T = ()
                        LB[()] += l[tr]

                if T:
                    LA[T] += l[tr]
            # if not tr:
            #     LA[()] += l[tr]

    return LA,LB  # new_logs is a list that contains logs






import pm4py

def conformance_checking(log_path, model_path):
    log = pm4py.read_xes(str(log_path))
    net,im,fm= pm4py.read.read_pnml(model_path)

    fitness_dict = pm4py.conformance.fitness_alignments(log,net,im,fm)
    precision = pm4py.conformance.precision_alignments(log,net,im,fm)

    return round(fitness_dict["log_fitness"],2),round(precision,2)

def extract_significant_dev(dev_list):
    list_dev = []
    for x in dev_list:
        if isinstance(x[1],str):
            list_dev.append((f"{x[0]}({x[1]})",x[2],x[3]))
        else:
            list_dev.append((f"{x[0]}({x[1][0]},{x[1][1]})",x[2],x[3]))
    sorted_IMr = sorted(list_dev, key=lambda x: x[1],reverse=True)
    return sorted_IMr[0:10]




import pm4py
import pandas as pd

def import_log(address):
    case_id_name = 'case:concept:name'
    timestamp_name = 'time:timestamp'
    acyivity_name = 'concept:name'

    log = pm4py.read_xes(str(address), variant="rustxes")
    case_table, event_table, map_info = log_to_tables(log, parameters={'case_id': case_id_name,
                                                                                        'timestamp': timestamp_name,
                                                                                        'activity_name': acyivity_name})
    case_table.to_csv('output_files/out.csv', index=False)
    event_table.to_csv('output_files/out_event.csv', index=False)


    return len(case_table),case_table.select_dtypes(include=['number']).columns


def log_to_tables(df, parameters):

    # Extract parameters
    case_id_name = parameters['case_id']
    timestamp_name = parameters['timestamp']
    activity_name = parameters['activity_name']

    # Constants
    time_unit = 24 * 3600  # Time unit in seconds (1 day)

    # Output column names
    output_case_id_name = 'case_id'
    output_timestamp_name = 'timestamp'
    output_activity_name = 'activity_name'

    # Identify case attributes (columns with unique values per case)
    dfunique = df.groupby(case_id_name).nunique()
    case_attributes = set(dfunique.columns[dfunique.max() <= 1])

    # Build the case table
    grouped = df.groupby(case_id_name)
    case_table = grouped.first()[list(case_attributes)]
    case_table['trace'] = grouped[activity_name].agg(lambda x: tuple(x.astype(str)))
    case_table['duration'] = (grouped[timestamp_name].max() - grouped[
        timestamp_name].min()).dt.total_seconds() // time_unit
    case_table['n_events'] = grouped.size()
    case_table = case_table.reset_index().rename(columns={case_id_name: output_case_id_name})

    # Build the event table
    non_attributes = {'row_num'}
    event_attributes = set(df.columns) - case_attributes - non_attributes
    event_table = df[list(event_attributes)]
    event_table[timestamp_name] = pd.to_datetime(event_table[timestamp_name], utc=True)

    # Sort events by timestamp (and EventOrder if available)
    sort_columns = [timestamp_name]
    if 'EventOrder' in event_table.columns:
        sort_columns.append('EventOrder')
    event_table = event_table.sort_values(sort_columns)

    # Rename columns for the event table
    event_table = event_table.rename(columns={
        case_id_name: output_case_id_name,
        timestamp_name: output_timestamp_name,
        activity_name: output_activity_name
    })

    # Mapping for standardized column names
    mapping = {
        'case_id': output_case_id_name,
        'timestamp': output_timestamp_name,
        'activity_name': output_activity_name
    }

    return case_table, event_table, mapping


import prolysis.rules_handling.declare_processing as declare_processing


def is_allowed(SS1,SS2,rules,low_pen,to_be_checked):
    dfa_intersect = rules[0]

    da_list = rules[1]
    S_mapped = rules[2]
    exclude = []
    S1 = {rules[2][x] for x in SS1}
    S2 = {rules[2][x] for x in SS2}
    penalty = {'seq': 0, 'exc': 0, 'par': 0, 'loop': 0, 'loop_tau': 0, 'exc_tau':0}
    for cut_type in to_be_checked:
        # if (declare_processing.check_all(S1, S2, dfa_intersect, cut_type, S_mapped)):
        # if (
        # declare_processing.check_all(S1, S2, S1 , S2 , dfa_intersect, cut_type,
        #                              S_mapped)):
        #     # print(dfa[0])
        #     exclude.append(cut_type)
        #     penalty[cut_type] += +1
        #
        for dfa in da_list:
            if (declare_processing.check_all(S1,S2,S1&{x for x in dfa[0][1]}, S2&{x for x in dfa[0][1]}, dfa[1], cut_type, S_mapped)):

                # print(dfa[0])
                exclude.append(cut_type)
                if penalty[cut_type]+dfa[2]>low_pen:
                    penalty[cut_type] += dfa[2]
                    break
                else:
                    penalty[cut_type] += dfa[2]
                    low_pen = min(low_pen,penalty[cut_type])
        if penalty[cut_type]==0:
            low_pen = 0

    # if sum(penalty.values())>0:
    #     print(penalty)
    # print(low_pen)
    return set(exclude), set(), penalty,low_pen


def is_allowed_single(activity, rules):
    penalty = {'single_single': 0, 'xor_single': 0, 'loop_single': 0}

    da_list = rules[1]
    S_mapped = rules[2]
    exclude = []

    for cut_type in {'single_single', 'xor_single', 'loop_single'}:
        for dfa in da_list:
            # if ("Absence" in dfa[0][0] or "Init" in dfa[0][0] or "End" in dfa[0][0] or "AtMost" in dfa[0][0] or "AtLeast" in dfa[0][0]):
            if dfa[0][1]==S_mapped[activity]:
                if (declare_processing.check_all_single(S_mapped[activity], dfa[1],
                                                 cut_type, S_mapped)):
                    print(dfa[0])
                    exclude.append(cut_type)
                    penalty[cut_type] += dfa[2]

    return set(exclude), penalty


import os
import subprocess

def mine_minerful_for_declare_constraints(window_size,sliding_window_size):
    # input_log_path = os.getcwd() + r"/output_files/log_ordered.xes"
    # output_log_path = os.getcwd()+ r"/output_files/behavioral_signals.csv"
    input_log_path = r"output_files/log_ordered.xes"
    output_log_path = r"output_files/behavioral_signals.csv"
    env = dict(os.environ)
    env = dict(os.environ)
    env['JAVA_OPTS'] = 'foo'
    subprocess.call(['java', '-version'])
    file_input = input_log_path
    subprocess.call([
        'java', "-Xmx16G",
        '-cp', f'MINERful.jar',
        'minerful.MinerFulMinerSlider',
        "-iLF", file_input,
        "-iLStartAt", "0",
        "-iLSubLen", str(window_size),
        "-sliBy", str(sliding_window_size),
        '-para', '4',
        '-s', '0.0',
        '-c', '0.0',
        # '-g', '0.0',
        '-prune', 'none',
        '-sliOut', output_log_path,
        '-vShush'
    ], env=env
        , cwd=os.getcwd())
# , cwd=r"/")

def prune_constraints_minerful(output_constraint_path,output_constraint_path_pruned):
    env = dict(os.environ)
    env['JAVA_OPTS'] = 'foo'
    subprocess.call(['java', "-Xmx16G", '-cp', f'MINERful.jar',
                     'minerful.MinerFulSimplificationStarter',
                     "-iSF",
                     output_constraint_path,
                     "-iSE", 'json',
                     "-oCSV", output_constraint_path_pruned,
                     "-keep",
                     "-prune", "hierarchyconflictredundancy"],
                    env=env,
                    cwd=os.getcwd())

def discover_declare(input_log_path,output_log_path,support, confidence):
    env = dict(os.environ)
    env['JAVA_OPTS'] = 'foo'
    subprocess.call(['java', '-version'])
    file_input = input_log_path
    subprocess.call([
        'java', "-Xmx16G",
        '-cp', f'MINERful.jar',
        'minerful.MinerFulMinerStarter',
        "-iLF", file_input,
        "-s", str(support),
        "-c", str(confidence),
        "-g", "0.0",
        "-sT", "0.00",
        "-cT", "0.00",
        "-gT", "0.0",
        '-prune', 'hierarchy',
        '-oJSON', output_log_path
    ], env=env
        , cwd=os.getcwd())


from setuptools import setup, find_packages

setup(
    name="prolysis",
    license="CC BY-NC 4.0",
    version="0.2.1",
    packages=find_packages(),
    install_requires=[
    "automata-lib ==8.4.0",
    "frozendict==2.4.4",
    "networkx==2.8.8",
    "matplotlib==3.10.0",
    "pm4py==2.7.13.1",
    "numpy==1.26.4",
    "pandas==2.2.3",
    "ruptures==1.1.9",
    "scikit_learn==1.6.0",
    "scipy==1.14.1",
    "seaborn==0.13.2",
    "pyemd==1.0.0",
    "redis==5.2.1",
    "numba==0.61.0",
    "pot==0.9.5",
    "rustxes==0.2.6"
    ],
    author="Ali Norouzifar",
    author_email="ali.norouzifar@pads.rwth-aachen.de",
    description="A Python package for process mining and analysis",
    url="https://github.com/aliNorouzifar/prolysis",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)


from typing import Optional, Dict, Any, Tuple
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.conversion.process_tree import converter as tree_to_petri
from pm4py.objects.process_tree.utils import generic
from pm4py.objects.process_tree.utils.generic import tree_sort
from prolysis.discovery.subtree_plain import SubtreePlain
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.process_tree.obj import Operator
from pm4py.util import constants
from enum import Enum
from collections import Counter
import json
import pandas as pd
import pm4py
from automata.fa.dfa import DFA
from pm4py.visualization.petri_net import visualizer as pn_visualizer
from pm4py.objects.petri_net.utils.reachability_graph import construct_reachability_graph
import ast
import os
import time
from pathlib import Path
from prolysis.rules_handling.utils import rules_from_json, preprocess, dfa_list_generator
import prolysis.rules_handling.declare_processing as declare_processing
from collections import Counter


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY



def apply_bi(Lp=pd.DataFrame(), Lm=pd.DataFrame(), parameters: Optional[Dict[Any, Any]] = None, sup= None, ratio = None, noise_thr =None, size_par = None, rules =None) -> Tuple[PetriNet, Marking, Marking]:
    file_path = r'output_files\discovery_log.json'
    with open(file_path, 'w') as file:
        json.dump([], file, indent=4)
    process_tree = apply_tree(Lp, Lm, parameters, sup=sup, ratio=ratio, noise_thr=noise_thr, size_par=size_par, rules=rules)
    net, initial_marking, final_marking = tree_to_petri.apply(process_tree)

    return net, initial_marking, final_marking



def apply_tree(logp,logm, parameters=None, sup= None, ratio = None, noise_thr =None, size_par = None, rules= None):
    if parameters is None:
        parameters = {}
    contains_empty_traces = False
    traces_length = [len(trace) for trace in logp]
    if traces_length:
        contains_empty_traces = min([len(trace) for trace in logp]) == 0

    recursion_depth = 0

    # logP_var = Counter(tuple([x['concept:name'] for x in t]) for t in logp)
    # logP_var = pm4py.stats.get_variants(logp)
    #
    # logM_var = pm4py.stats.get_variants(logm)
    if logm.empty:
        logm_var = Counter()
    else:
        logm_var = pm4py.stats.get_variants(logm)
    logp_var = pm4py.stats.get_variants(logp)
    sub = SubtreePlain(logp_var,logm_var, recursion_depth, noise_thr, parameters=parameters, sup= sup, ratio = ratio, size_par = size_par, rules= rules)

    process_tree = get_repr(sub, 0, contains_empty_traces=contains_empty_traces)
    # Ensures consistency to the parent pointers in the process tree
    fix_parent_pointers(process_tree)
    # Fixes a 1 child XOR that is added when single-activities flowers are found
    fix_one_child_xor_flower(process_tree)
    # folds the process tree (to simplify it in case fallthroughs/filtering is applied)
    process_tree = generic.fold(process_tree)
    # sorts the process tree to ensure consistency in different executions of the algorithm
    tree_sort(process_tree)

    return process_tree



def get_repr(spec_tree_struct, rec_depth, contains_empty_traces=False):



    base_cases = ('empty_log', 'single_activity')
    cut = ('xor', 'sequential', 'parallel', 'loopCut')


    # if a cut was detected in the current subtree:
    if spec_tree_struct.detected_cut in cut:
        if spec_tree_struct.detected_cut == "sequential":
            final_tree_repr = ProcessTree(operator=Operator.SEQUENCE)
        elif spec_tree_struct.detected_cut == "loopCut":
            final_tree_repr = ProcessTree(operator=Operator.LOOP)
        elif spec_tree_struct.detected_cut == "xor":
            final_tree_repr = ProcessTree(operator=Operator.XOR)
        elif spec_tree_struct.detected_cut == "parallel":
            final_tree_repr = ProcessTree(operator=Operator.PARALLEL)

        if not (spec_tree_struct.detected_cut == "loopCut" and len(spec_tree_struct.children) >= 3):
            for ch in spec_tree_struct.children:
                # get the representation of the current child (from children in the subtree-structure):
                child = get_repr(ch, rec_depth + 1)
                # add connection from child_tree to child_final and the other way around:
                final_tree_repr.children.append(child)
                child.parent = final_tree_repr

        else:
            child = get_repr(spec_tree_struct.children[0], rec_depth + 1)
            final_tree_repr.children.append(child)
            child.parent = final_tree_repr

            redo_child = ProcessTree(operator=Operator.XOR)
            for ch in spec_tree_struct.children[1:]:
                child = get_repr(ch, rec_depth + 1)
                redo_child.children.append(child)
                child.parent = redo_child

            final_tree_repr.children.append(redo_child)
            redo_child.parent = final_tree_repr

        if spec_tree_struct.detected_cut == "loopCut" and len(spec_tree_struct.children) < 3:
            while len(spec_tree_struct.children) < 2:
                child = ProcessTree()
                final_tree_repr.children.append(child)
                child.parent = final_tree_repr
                spec_tree_struct.children.append(None)

    if spec_tree_struct.detected_cut in base_cases:
        # in the base case of an empty log, we only return a silent transition
        if spec_tree_struct.detected_cut == "empty_log":
            return ProcessTree(operator=None, label=None)
        # in the base case of a single activity, we return a tree consisting of the single activity
        elif spec_tree_struct.detected_cut == "single_activity":
            act_a = spec_tree_struct.activitiesP.pop()
            return ProcessTree(operator=None, label=act_a)
    return final_tree_repr


def get_transition(label):
    """
    Create a node (transition) with the specified label in the process tree
    """
    return ProcessTree(operator=None, label=label)

class Counts(object):
    """
    Shared variables among executions
    """

    def __init__(self):
        """
        Constructor
        """
        self.num_places = 0
        self.num_hidden = 0
        self.num_visible_trans = 0
        self.dict_skips = {}
        self.dict_loops = {}

    def inc_places(self):
        """
        Increase the number of places
        """
        self.num_places = self.num_places + 1

    def inc_no_hidden(self):
        """
        Increase the number of hidden transitions
        """
        self.num_hidden = self.num_hidden + 1

    def inc_no_visible(self):
        """
        Increase the number of visible transitions
        """
        self.num_visible_trans = self.num_visible_trans + 1


def fix_parent_pointers(pt):
    """
    Ensures consistency to the parent pointers in the process tree

    Parameters
    --------------
    pt
        Process tree
    """
    for child in pt.children:
        child.parent = pt
        if child.children:
            fix_parent_pointers(child)


def fix_one_child_xor_flower(tree):
    """
    Fixes a 1 child XOR that is added when single-activities flowers are found

    Parameters
    --------------
    tree
        Process tree
    """
    if tree.parent is not None and tree.operator is Operator.XOR and len(tree.children) == 1:
        for child in tree.children:
            child.parent = tree.parent
            tree.parent.children.append(child)
            del tree.parent.children[tree.parent.children.index(tree)]
    else:
        for child in tree.children:
            fix_one_child_xor_flower(child)


def run_IMr(LPlus_LogFile,support,rules, activities,dim,abs_thr):
    event_log_xes = pm4py.read_xes(str(LPlus_LogFile), variant="rustxes")
    if rules !=[]:
        activities = set(activities)
        lookup_table_path = Path("assets") / "lookup_table.csv"
    
        # if rules_path=="":
        #     rules= {}
        #     activities = {}
        # else:
        #     rules, activities = rules_from_json(str(rules_path))
        rules_proccessed, absence_list = preprocess(rules, dim, abs_thr)

        event_log_xes = pm4py.filter_event_attribute_values(
            event_log_xes,
            attribute_key="concept:name",  # Default attribute for activity names in XES
            values=absence_list,
            retain=False  # Retain=False means we exclude the specified activities
        )

        print('rules are preprocessed')

        lookup_table = pd.read_csv(lookup_table_path, sep=';', index_col=0)
        lookup_table = lookup_table.map(lambda x: ast.literal_eval(x) if isinstance(x, str) else x)

        S_mapped = declare_processing.assign_alphabet(activities)

        print('conversion of rules to DFAs started')
        dfa_list, total_sup, total_conf = dfa_list_generator(rules_proccessed, S_mapped)
        print('conversion of rules to DFAs ended')

        print(f"_______________ support is {support}_____________________")
        print('process discovery started')
        start = time.time()
        net, initial_marking, final_marking = apply_bi(Lp=event_log_xes, sup=support, ratio=0, size_par=1,
                                                                 rules=(rules_proccessed, lookup_table, dim))
        end = time.time()
        print(end - start)
        print('process discovery ended')

        pm4py.write_pnml(net, initial_marking, final_marking, os.path.join(r"output_files", "model.pnml"))



        print('model_checking started')
        rg = construct_reachability_graph(net, initial_marking, use_trans_name=False, parameters=None)
        aa = declare_processing.reachability2NFA(rg, activities)
        model_dfa = DFA.from_nfa(aa)
        cond = []
        sup_cost = 0
        conf_cost = 0
        for dfa in dfa_list:
            constraint_dfa_complement = dfa[1].complement()
            intersection_dfa = model_dfa.intersection(constraint_dfa_complement)
            if not intersection_dfa.isempty():
                sup_cost += dfa[2]
                conf_cost += dfa[2]
            cond.append((dfa[0], intersection_dfa.isempty(), dfa[2], dfa[3]))
        print('model_checking ended')

        report = {}
        report['time'] = end - start
        report['N.rules'] = len(dfa_list)
        report['N.dev'] = len([(x[0][0], x[0][1]) for x in cond if x[1] == False])
        if total_sup!=0:
            report['support_cost'] = round(sup_cost / total_sup, 2)
        else:
            report['support_cost'] = 0
        if total_conf != 0:
            report['confidence_cost'] = round(conf_cost / total_conf, 2)
            report['confidence_cost'] = 0
        report['dev_list'] = [(x[0][0], x[0][1], round(x[2], 2), round(x[3], 2)) for x in cond if x[1] == False]

        with open(os.path.join(r"output_files/", "stats.json"), "w") as json_file:
            json.dump(report, json_file, indent=4)

        print('The report is generated')
    else:
        net, initial_marking, final_marking = apply_bi(Lp=event_log_xes, sup=support, ratio=0, size_par=1,
                                                       rules=([], [], ""))


    parameters = {pn_visualizer.Variants.WO_DECORATION.value.Parameters.FORMAT: "png"}
    gviz = pn_visualizer.apply(net, initial_marking, final_marking, parameters=parameters)
    gviz.attr("graph", bgcolor="transparent")
    return gviz



def run_IMr_norule(LPlus_LogFile,support):

    event_log_xes = pm4py.read_xes(LPlus_LogFile, variant="rustxes")

    print(f"_______________ support is {support}_____________________")
    print('process discovery started')
    start = time.time()
    net, initial_marking, final_marking = apply_bi(Lp=event_log_xes, sup=support, ratio=0, size_par=1,
                                                             rules=({}, {}))
    end = time.time()
    print(end - start)
    print('process discovery ended')

    pm4py.write_pnml(net, initial_marking, final_marking, os.path.join(r"output_files", "model.pnml"))



    return net, initial_marking, final_marking



import prolysis.rules_handling.declare_processing as declare_processing


def is_allowed(SS1,SS2,rules_lookup):
    dim = 'support'
    rules =  rules_lookup[0]
    lookup_table = rules_lookup[1]
    SS = SS1.union(SS2)

    exclude = set()
    penalty = {'seq': 0, 'exc': 0, 'par': 0, 'loop': 0, 'loop_tau': 0, 'exc_tau': 0}
    for r in rules:
        x =r["template"]
        if len(r['parameters'])==2:
            if (r['parameters'][0][0] not in SS) or (r['parameters'][1][0] not in SS):
                continue
            if SS2:
                y = f"{int(r['parameters'][0][0] in SS1)}{int(r['parameters'][1][0] in SS1)}00"
            else:
                y = f"{int(r['parameters'][0][0] in SS1)}{int(r['parameters'][1][0] in SS1)}10"
        else:
            if (r['parameters'][0][0] not in SS):
                continue
            else:
                if SS2:
                    y = f"{int(r['parameters'][0][0] in SS1)}000"
                else:
                    y = f"{int(r['parameters'][0][0] in SS1)}110"
        exclude_new = lookup_table.loc[x][y]
        # if not SS2 and ('loop_tau' in exclude_new):
        #     print('wait')
        for ex in exclude_new:
            penalty[ex] += r[dim]
        exclude = exclude | exclude_new
        # if penalty>=low_pen:
        #     break
    return exclude, set(), penalty



def is_allowed_single(activity,rules_lookup):
    rules =  rules_lookup[0]
    lookup_table = rules_lookup[1]
    dim = 'support'
    exclude = set()
    penalty = {'single_single': 0, 'xor_single': 0, 'loop_single': 0}
    for r in rules:
        if len(r['parameters']) == 2 or activity!=r['parameters'][0][0]:
            continue
        elif activity==r['parameters'][0][0]:
            x =r["template"]
            y = f"0001"
            exclude_new = lookup_table.loc[x][y]
            for ex in exclude_new:
                penalty[ex] += r[dim]
            exclude = exclude | exclude_new
            # if penalty>=low_pen:
            #     break
    return exclude, penalty




from ast import literal_eval
from collections.abc import Iterable
from dataclasses import dataclass
import dataclasses
import functools
import json
from multiprocessing import Pool
import time
import logging

from collections import defaultdict
from itertools import chain
from pathlib import Path
from typing import Dict, Tuple
import argparse

from pandas import DataFrame
import pm4py
from pm4py.objects.log.obj import EventLog
from pm4py.objects.log.importer.xes import importer as xes_importer
from pm4py.objects.petri_net.importer import importer as pnml_importer
from prolysis.analysis.python_emsc import stochastic as stc
from prolysis.analysis.python_emsc.util import snake_to_camel

logger = logging.getLogger(__name__)

def _process_petri_net(df_ev, net, im):
    # Consider all activities not contained in log silent
    # (Current PM4Py import makes ProM tau transitions visible)
    for t in net.transitions:
        if t.properties['stochastic_distribution'].random_variable.weight == 0:
            logger.warning(f"Transition {str(t)} has weight zero -> set this to a very small value")
            t.properties['stochastic_distribution'].random_variable.weight = 0.0000001


    # No initial marking found
    # Set initial marking assuming that the Petri net
    # is a workflow net
    if len(im) == 0:
        logger.warning("Initial marking of SPN is empty. Creating one under workflow net assumption.")
        for p in net.places:
            # No incoming arcs
            if len(p.in_arcs) == 0:
                im[p] += 1
        logger.warning(f"Created initial marking {str(im)}")


def time_callback(callback, *args, **kwargs):
    start_time = time.perf_counter()
    value = callback(*args, **kwargs)
    end_time = time.perf_counter()
    run_time = end_time - start_time
    return value, run_time


def timer(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return time_callback(func, *args, **kwargs)

    return wrapper


DATASETS_SCRIPT_FOLDER = Path(__file__).parent.resolve()

DATA_FOLDER = DATASETS_SCRIPT_FOLDER / 'data'
MODELS_FOLDER = DATA_FOLDER / 'models'
LOGS_FOLDER = DATA_FOLDER / 'logs'

@dataclass
class MySPNEvalResult:
    log_name: str
    name_spn: str
    nbr_path: int
    sampled_prob: float
    emsc: float
    pemsc: float

    def to_json(self, include_null=False) -> dict:
        """Converts this to json. Assumes variables are snake cased, converts to camel case.

        Args:
            include_null (bool, optional): Whether null values are included. Defaults to False.

        Returns:
            dict: Json dictionary
        """
        return dataclasses.asdict(
            self,
            dict_factory=lambda fields: {
                snake_to_camel(key): value
                for (key, value) in fields
                if value is not None or include_null
            },
        )


def load_log(log_name: str):
    # Tries XES first, else CSV
    log_path_xes = LOGS_FOLDER / log_name / (log_name + ".xes")

    variant = xes_importer.Variants.ITERPARSE
    parameters = {variant.value.Parameters.TIMESTAMP_SORT: True}
    return xes_importer.apply(str(log_path_xes),
                              variant=variant, parameters=parameters)


def read_gspn_file(model: str, variant: str = '') -> stc.StochasticPetriNet:
    model_path = MODELS_FOLDER / model / (model + f'{variant}' + ".pnml")
    wf_net = pm4py.read_pnml(str(model_path), auto_guess_final_marking=True)

    # ProM exports the model wrongly.
    # We must set the taus and estimate the initial marking
    for transition in wf_net[0].transitions:
        if transition.label is not None:
            if 'tau' in transition.label or transition.label in {'start', 'end'}:
                transition.label = None

    if len(wf_net[1]) == 0:
        for place in wf_net[0].places:
            if len(place.in_arcs) == 0:
                wf_net[1][place] = 1

    return get_stochastic_information(wf_net)


def get_stochastic_information(wf_net: stc.WFNet) -> stc.StochasticPetriNet:
    net, im, fm = wf_net

    weights = {}
    for transition in net.transitions:
        stochastic_information = transition.properties['stochastic_distribution']
        weights[transition] = stochastic_information.random_variable.weight

    return stc.StochasticPetriNet(wf_net=wf_net, weights=weights)


def deduplicate_log_labels(log: EventLog):
    # TODO fix this hack
    alphabet = 'abcdefghijklmnopqrstuvxwyz'
    digits = '0123456789'
    tokens = chain(alphabet, alphabet.upper(), digits, [None])
    label_map = defaultdict(lambda: next(tokens))

    for trace in log:
        for event in trace:
            activity = event['concept:name']
            event['concept:name'] = label_map[activity]

    return log, label_map


@timer
def init_model_language(model, max_num_paths=15_000):
    path_language = stc.sample_most_likely_net_paths(model, 0.99, 15_000, 100_000)
    sampled_language = stc.path_language_to_trace_language(path_language)
    return path_language, sampled_language
    

@timer
def compute_emsc(stochastic_log_language, truncated_model_language, emsc_type: stc.EMSCType=stc.EMSCType.EMSC) -> Tuple[float, int, float]:
    try:
        emd = stc.compare_languages_levenshtein(stochastic_log_language, truncated_model_language, emsc_type=emsc_type)
        if emd < 0:
            return -1
        else:
            return 1 - emd
    except Exception as e:
        logger.error(e)
        return -1


def compute_my_metrics(stochastic_log_language, model, max_num_paths=15_000):
    (path_language, truncated_model_language), time_sampling = init_model_language(model, max_num_paths=max_num_paths)
    sampled_prob = sum(prob for _, prob in truncated_model_language.items())
    # Original EMSC
    emsc, time_emsc = compute_emsc(stochastic_log_language, truncated_model_language, emsc_type=stc.EMSCType.EMSC)
    # 1-Penalized EMSC
    pemsc, time_pemsc = compute_emsc(stochastic_log_language, truncated_model_language, emsc_type=stc.EMSCType.PEMSC)
    return emsc, pemsc, len(path_language), sampled_prob


def list_log_path_pn_directory_type(s):
    try:
        log_pn_dir_pairs = literal_eval(s)
    except:
        print(s)
        raise argparse.ArgumentTypeError("Cannot Evaluate (ast.literal_eval): (Log name, Log path, SPN directory), (...), ... should be a enumeration of triples that can be parsed as Python code")
    # Check enumeration of tuples
    if not isinstance(log_pn_dir_pairs, Iterable):
        raise argparse.ArgumentTypeError("Not an enumeration! (Log name, Log path, SPN directory), (...), ... should be a enumeration of triples.")
    if not all(isinstance(item, tuple) for item in log_pn_dir_pairs):
        raise argparse.ArgumentTypeError("Not an enumeration of tuples! (Log name, Log path, SPN directory), (...), ... should be a enumeration of triples.")
    if not all(len(item) == 3 for item in log_pn_dir_pairs):
        raise argparse.ArgumentTypeError("Not an enumeration of triples! (Log name, Log path, SPN directory), (...), ... should be a enumeration of triples.")
    if not all(isinstance(log_name, str) and isinstance(log_path, str) and isinstance(spn_directory, str) for  (log_name, log_path, spn_directory) in log_pn_dir_pairs):
        raise argparse.ArgumentTypeError("Not an enumeration of string triples! (Log name, Log path, SPN directory), (...), ... should be a enumeration of triples.")
    
    log_pn_dir_pairs_as_paths = tuple((log_name, Path(log_path_str), Path(spn_directory)) for (log_name, log_path_str, spn_directory) in log_pn_dir_pairs)
    return log_pn_dir_pairs_as_paths


def _main_emscc_ot_evaluation_full_cmd():
    logging.basicConfig(level=logging.INFO)
    ##############################
    # Arguments
    ##############################
    parser = argparse.ArgumentParser(
        prog='EMSCC on multiple log-SPN combinations',
        description=''
    )

    parser.add_argument('logSPNPairs', help='String pairs (log path, SPN directory) that can be evaluated using ast.literal_eval', 
                        type=list_log_path_pn_directory_type)
    parser.add_argument('--resultFile', type=str, help='Result path must be a json file (directory must exist)')
    parser.add_argument('--poolSize', type=int, help='Pool Size used for concurrent computation', default=10)

    args = parser.parse_args()
    log_SPN_pairs = args.logSPNPairs
    path_result = Path(args.resultFile)
    pool_size = args.poolSize

    logger.info(f"Config read from command line: {str(log_SPN_pairs)}")
    # Check if input files and directories exist
    for (log_name, path_log, path_spn_dir) in log_SPN_pairs:
        if not path_log.is_file():
            raise Exception(f'Event log does not exist: {str(path_log)}')
        if not path_spn_dir.is_dir():
            raise Exception(f'SPN directory does not exist: {str(path_spn_dir)}')

    # Write result header
    with open(path_result, 'a') as f:
        f.write("LogName,SPNName,EMSCC,EMSCCTime,TimeWasserstein,TimeOT,SampledProb,NbrPaths\n")

    emsc_spn_results_full = []
    for log_name, path_log, path_spn_dir in log_SPN_pairs:
        # Load log
        logger.info(f"Importing log {path_log}")
        log = pm4py.read_xes(str(path_log))
        df_ev = pm4py.convert_to_dataframe(log)

        # Stochastic log language
        logger.info(f"Creating stochast language for log {path_log}")
        stochastic_log_language = stc.log_to_stochastic_language(log)


        ##############################
        # Evaluate EMSCC for all SPNs in directory 
        ##############################
        logger.info(f"Start Parallel EMSC Computation with {pool_size} processes")
        # Iterate over Petri nets
        run_spn_evaluation_on_log_partial = functools.partial(run_spn_evaluation_on_log, log_name=log_name, df_ev=df_ev, stochastic_log_language=stochastic_log_language)
        with Pool(processes=pool_size) as pool:
            #emsc_spn_results = pool.map(run_spn_evaluation_on_log_partial, itertools.islice(path_spn_dir.glob('*.pnml'), 3))
            emsc_spn_results = pool.map(run_spn_evaluation_on_log_partial, path_spn_dir.glob('*.pnml'))
        emsc_spn_results_full +=  emsc_spn_results

    # Write results
    data = { 'SPNResultEMSC': list(map(MySPNEvalResult.to_json, emsc_spn_results_full)) }
    with open(path_result, 'w') as f:
        json.dump(data, f, indent=4)


def run_spn_evaluation_on_log(path_pn: Path, log_name: str, df_ev: DataFrame, stochastic_log_language: Dict[Tuple[str], float]) -> MySPNEvalResult:
    net, im, fm, stochastic_map = pnml_importer.apply(str(path_pn), parameters={"return_stochastic_map": True})
    _process_petri_net(df_ev, net, im)
    spn = get_stochastic_information((net, im, fm))

    emsc, pemsc, nbr_path, sampled_prob = compute_my_metrics(stochastic_log_language, spn)
    if emsc < 0 or pemsc < 0:
        logger.error(f'Failed to compute EMSC or PEMS for {path_pn}')

    name_spn = path_pn.name.replace('.pnml', '')
    return MySPNEvalResult(log_name=log_name, name_spn=name_spn,
                           nbr_path=nbr_path, sampled_prob=sampled_prob,
                           emsc=emsc, pemsc=pemsc)


if __name__ == '__main__':
    _main_emscc_ot_evaluation_full_cmd()
    # TODO Distinguish between timed and immediate transitions
    #print(_main_emsc_impl())
